# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2023 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Base class for asynchronous processing
"""

import math
import os
import pickle
import requests
import shutil
import subprocess
import sys
import tempfile
import time
import traceback
import uuid

from flask import json
from requests.auth import HTTPBasicAuth

from actinia_core.core.common.config import global_config, DEFAULT_CONFIG_PATH
from actinia_core.core.common.process_object import Process
from actinia_core.core.grass_init import GrassInitializer
from actinia_core.core.messages_logger import MessageLogger
from actinia_core.core.redis_lock import RedisLockingInterface
from actinia_core.core.resources_logger import ResourceLogger
from actinia_core.core.mapset_merge_utils import change_mapsetname
from actinia_core.core.common.process_chain import ProcessChainConverter
from actinia_core.core.common.exceptions import (
    AsyncProcessError,
    AsyncProcessTermination,
    RsyncError,
)
from actinia_core.core.common.exceptions import AsyncProcessTimeLimit
from actinia_core.models.response_models import (
    ProcessingResponseModel,
    ExceptionTracebackModel,
)
from actinia_core.models.response_models import (
    create_response_from_model,
    ProcessLogModel,
    ProgressInfoModel,
)
from actinia_core.core.interim_results import InterimResult, get_directory_size
from actinia_core.rest.base.user_auth import (
    check_location_mapset_module_access,
)

__license__ = "GPLv3"
__author__ = "Sören Gebbert, Anika Weinmann, Lina Krisztian"
__copyright__ = (
    "Copyright 2016-2023, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis GmbH & Co. KG"


class EphemeralProcessing(object):
    """This class processes GRASS data on the local machine in a temporary
    mapset.

    The temporary mapset will be removed by this class when the processing is
    finished

    Creating the temporary database and mapset:

    1. Create a new gisdbase in a temporary directory

       e.g: /tmp/soeren_disdbase

    2. Create the required location directory

       e.g: /tmp/soeren_temp_gisdbase/ECAD

    3. Softlink the PERMANENT and all required mapsets into the
       new location directory from the original location,
       check the input parameter of the module for which mapsets must be linked

       e.g: /mount/groups/[user group]/locations/ECAD/PERMANENT
        -> /tmp/soeren_temp_gisdbase/ECAD/PERMANENT
       e.g: /mount/groups/[user group]/locations/ECAD/Temperature
        -> /tmp/soeren_temp_gisdbase/ECAD/Temperature

    4. Set the GRASS GIS environmental variables to point to the new gisdbase,
       location and PERMANENT mapset

    5. Create a new mapset with g.mapset in the temporary location directory

       e.g: /tmp/soeren_temp_gisdbase/ECAD/MyMapset

    6. Run and poll each process in the process chain until finished and update
       the resource log entry
       - Create a poll loop and check the subprocess status
       - Update the resource entry with progress info

    7. When finished create the response:
         * put the output of all modules into the response document

    8. Remove the temporary gisdbase

    9. Send the response document with status finished and resource links

    """

    def __init__(self, rdc):
        """Constructor

        Args:
            rdc (ResourceDataContainer): The data container that contains all
                                         required variables for processing

        """
        # Fluentd hack to work in a multiprocessing environment
        try:
            from fluent import sender

            self.has_fluent = True
        except Exception:
            self.has_fluent = False

        # rdc = ResourceDataContainer()

        self.rdc = rdc
        if os.path.exists(DEFAULT_CONFIG_PATH) is True and os.path.isfile(
            DEFAULT_CONFIG_PATH
        ):
            self.config = global_config
            self.rdc.config = self.config
        else:
            self.config = self.rdc.config

        self.data = self.rdc.user_data
        self.grass_temp_database = self.config.GRASS_TMP_DATABASE

        self.request_data = self.rdc.request_data
        self.user_id = self.rdc.user_id
        self.user_group = self.rdc.user_group
        self.user_credentials = rdc.user_credentials

        self.resource_id = self.rdc.resource_id
        self.iteration = self.rdc.iteration
        self.status_url = self.rdc.status_url
        self.api_info = self.rdc.api_info
        self.interim_result = InterimResult(
            self.user_id,
            self.resource_id,
            self.iteration,
            self.api_info["endpoint"],
        )

        self.grass_data_base = self.rdc.grass_data_base  # Global database
        # User database base path, this path will be
        self.grass_user_data_base = self.rdc.grass_user_data_base
        # extended with the user group name in the setup
        self.grass_base_dir = self.rdc.grass_base_dir

        self.location_name = self.rdc.location_name
        self.mapset_name = self.rdc.mapset_name
        # Set this True if the work is performed based on global database
        self.is_global_database = False

        self.map_name = self.rdc.map_name

        self.orig_time = self.rdc.orig_time
        self.orig_datetime = self.rdc.orig_datetime

        # Create the unique temporary gisdbase and mapset names
        self.unique_id = str(uuid.uuid4()).replace("-", "")
        self.temp_grass_data_base_name = "gisdbase_" + self.unique_id
        self.temp_mapset_name = "mapset_" + self.unique_id
        self.temp_mapset_path = None

        self.ginit = None

        # Successfully finished message
        self.finish_message = "Processing successfully finished"

        # The temporary GRASS database
        self.temp_grass_data_base = (
            None  # Path to the temporary grass database
        )
        self.temp_location_path = None  # Path to the temporary location
        self.temp_file_path = None  # The path to store temporary created files
        self.global_location_path = (
            None  # The path to the global location to link
        )
        self.user_location_path = None  # The path to the user location to link

        # List of resources that should be created
        self.resource_export_list = list()
        self.resource_url_list = list()

        # Initialize the user specific permissions
        self.cell_limit = 0
        self.process_time_limit = 0
        self.process_num_limit = 0
        # Set this True so that regions are not checked before processing
        self.skip_region_check = False

        # The stdout, stderr and parameter log of the module chains
        self.module_output_log = list()
        # The stdout, stderr and parameter log of the module chains
        # using a dict with the process id as key
        self.module_output_dict = dict()
        # The list of output parser definitions that must be applied
        # after the module run. The parser result will be stored in
        # the module_result dictionary using the parser id
        self.output_parser_list = []
        # A dictionary that has the process id as key to store module
        self.module_results = dict()
        # outputs like images, dicts, files and so on

        self.required_mapsets = (
            list()
        )  # The process chain analysis will provide
        # a list of required mapsets that must be
        # linked in the temporary location

        # The module that was called in the process chain, to detect g.region
        # calls and check for correct region settings
        self.last_module = "g.region"
        # Count the processes executed from the process chain
        self.process_count = 0

        self.ginit = None

        # The state of the whole processing that is checked in the final
        # section of the run function
        self.run_state = {"success": None}

        # The progress info object
        self.progress = ProgressInfoModel(step=0, num_of_steps=0)
        # The count of self._run_process() and self._run_module() calls
        self.progress_steps = 0
        # The number of processes that should be processes
        self.number_of_processes = 0

        self.setup_flag = False

        # The names of the temporarily generated files
        # "key":"temporary_file_path"
        self.temporary_pc_files = {}
        # The counter to generate unique temporary file names
        self.temp_file_count = 0
        # This dictionary contains the output of a process on stdout
        self.process_dict = {}

        # The class that is used to create the response
        self.response_model_class = ProcessingResponseModel
        # The class that converts process chain definitions into
        self.proc_chain_converter = None
        # process lists that will be executed. This variable is
        # initiated in the setup method
        # The list of all process chains that were processed
        self.process_chain_list = []
        # A list of all processes that will be executed
        self.actinia_process_list = list()
        # A list of all processes that will be executed
        self.actinia_process_dict = dict()
        # The URL of a webhook that should becalled after processing of a
        self.webhook_finished = None
        # process chain finished

        # The URL of a webhook that should be called for each status/progress
        # update
        self.webhook_update = None
        # The authentication for the webhook (base 64 decoded
        # "username:password")
        self.webhook_auth = None

    def _send_resource_update(self, message, results=None):
        """Create an HTTP response document and send it to the status database

        Args:
            message (str): The message
            results (dict): Results of the processing using the process chain
                            id for identification

        """
        data = create_response_from_model(
            self.response_model_class,
            status="running",
            user_id=self.user_id,
            resource_id=self.resource_id,
            queue=self.rdc.queue,
            iteration=self.iteration,
            # process_log=self.module_output_log,
            progress=self.progress,
            results=results,
            message=message,
            orig_time=self.orig_time,
            orig_datetime=self.orig_datetime,
            http_code=200,
            status_url=self.status_url,
            api_info=self.api_info,
        )
        self._send_to_database(document=data, final=False)

    def _send_resource_finished(self, message, results=None):
        """Create an HTTP response document and send it to the status database

        Args:
            message (str): The message
            results (dict): Results of the processing using the process chain
                            id for identification

        """
        data = create_response_from_model(
            self.response_model_class,
            status="finished",
            user_id=self.user_id,
            resource_id=self.resource_id,
            queue=self.rdc.queue,
            iteration=self.iteration,
            process_log=self.module_output_log,
            progress=self.progress,
            results=results,
            message=message,
            orig_time=self.orig_time,
            orig_datetime=self.orig_datetime,
            http_code=200,
            status_url=self.status_url,
            api_info=self.api_info,
            resource_urls=self.resource_url_list,
            process_chain_list=self.process_chain_list,
        )
        self._send_to_database(document=data, final=True)

    def _send_resource_terminated(self, message, results=None):
        """Create an HTTP response document and send it to the status database

        Args:
            message (str): The message
            results (dict): Results of the processing using the process chain
                            id for identification

        """
        data = create_response_from_model(
            self.response_model_class,
            status="terminated",
            user_id=self.user_id,
            resource_id=self.resource_id,
            queue=self.rdc.queue,
            iteration=self.iteration,
            process_log=self.module_output_log,
            progress=self.progress,
            results=results,
            message=message,
            orig_time=self.orig_time,
            orig_datetime=self.orig_datetime,
            http_code=200,
            status_url=self.status_url,
            api_info=self.api_info,
            process_chain_list=self.process_chain_list,
        )
        self._send_to_database(document=data, final=True)

    def _send_resource_time_limit_exceeded(self, message, results=None):
        """Create an HTTP response document and send it to the status database

        Args:
            message (str): The message
            results (dict): Results of the processing using the process chain
                            id for identification

        """
        data = create_response_from_model(
            self.response_model_class,
            status="terminated",
            user_id=self.user_id,
            resource_id=self.resource_id,
            queue=self.rdc.queue,
            iteration=self.iteration,
            process_log=self.module_output_log,
            progress=self.progress,
            results=results,
            message=message,
            orig_time=self.orig_time,
            orig_datetime=self.orig_datetime,
            http_code=400,
            status_url=self.status_url,
            api_info=self.api_info,
            process_chain_list=self.process_chain_list,
        )
        self._send_to_database(document=data, final=True)

    def _send_resource_error(self, message, results=None, exception=None):
        """Create an HTTP response document and send it to the status database

        Args:
            message (str): The message
            results (dict): Results of the processing using the process chain
                            id for identification

        """
        data = create_response_from_model(
            self.response_model_class,
            status="error",
            user_id=self.user_id,
            resource_id=self.resource_id,
            queue=self.rdc.queue,
            iteration=self.iteration,
            process_log=self.module_output_log,
            progress=self.progress,
            results=results,
            message=message,
            orig_time=self.orig_time,
            orig_datetime=self.orig_datetime,
            http_code=400,
            status_url=self.status_url,
            api_info=self.api_info,
            process_chain_list=self.process_chain_list,
            exception=exception,
        )
        self._send_to_database(document=data, final=True)

    def _send_to_database(self, document, final=False):
        """Send the document to the database

        The resource expiration time set in the actinia config file will be
        used for every resource commit.

        If a webhook URL is provided, the JSON response will be send to the
        provided endpoint using a POST request.

        Args:
            document (str): The response document
            final (bool): Set True if this was the final resource commit
                          (no update) to activate the webhook call

        """

        self.resource_logger.commit(
            user_id=self.user_id,
            resource_id=self.resource_id,
            iteration=self.iteration,
            document=document,
            expiration=self.config.REDIS_RESOURCE_EXPIRE_TIME,
        )

        # Call the webhook after the final result was send to the database
        try:
            if final is True and self.webhook_finished is not None:
                self._post_to_webhook(document, "finished")
            elif final is False and self.webhook_update is not None:
                self._post_to_webhook(document, "update")
        except Exception as e:
            e_type, e_value, e_tb = sys.exc_info()
            model = ExceptionTracebackModel(
                message=str(e_value),
                traceback=traceback.format_tb(e_tb),
                type=str(e_type),
            )
            run_state = {"error": str(e), "exception": model}
            print(str(run_state))
            self.message_logger.error(
                "Unable to send webhook request. Traceback: %s"
                % str(run_state)
            )

    def _post_to_webhook(self, document, type):
        """Helper method to send a post request to a webhook.
        The finished webhook will be retried until it has reached the number
        of tries. The number of tries is WEBHOOK_RETRIES which can be set in the
        config.

        Args:
            document (str): The response document
            type (str): The webhook type: 'finished' or 'update'
        """
        self.message_logger.info(
            "Send POST request to %s webhook url: %s"
            % (type, self.webhook_finished)
        )
        webhook_url = None
        if type == "finished":
            webhook_url = self.webhook_finished
            webhook_retries = self.config.WEBHOOK_RETRIES
            webhook_sleep = self.config.WEBHOOK_SLEEP
        if type == "update":
            webhook_url = self.webhook_update
            webhook_retries = 1
            webhook_sleep = 0

        http_code, response_model = pickle.loads(document)

        webhook_not_reached = True
        retry = 0
        while webhook_not_reached is True and retry < webhook_retries:
            retry += 1
            try:
                if self.webhook_auth:
                    # username is expected to be without colon (':')
                    resp = requests.post(
                        webhook_url,
                        json=json.dumps(response_model),
                        auth=HTTPBasicAuth(
                            self.webhook_auth.split(":")[0],
                            ":".join(self.webhook_auth.split(":")[1:]),
                        ),
                        timeout=10,
                    )
                else:
                    resp = requests.post(
                        webhook_url,
                        json=json.dumps(response_model),
                        timeout=10,
                    )
                if not (500 <= resp.status_code and resp.status_code < 600):
                    webhook_not_reached = False
            except Exception:
                time.sleep(webhook_sleep)
        if (
            webhook_not_reached is False and resp.status_code not in [200, 204]
        ) or webhook_not_reached is True:
            raise AsyncProcessError(
                "Unable to access %s webhook URL %s" % (type, webhook_url)
            )

    def _get_previous_iteration_process_chain(self):
        """Helper method to check the old resource run and get the step of the
        process chain where to continue

        Returns:
            pc_step (int): The number of the step in the process chain where to
                           continue
           old_process_chain (dict): The process chain of the old resource run
        """
        # check old resource
        pc_step = 0

        for iter in range(1, self.rdc.iteration):
            if iter == 1:
                old_response_data = self.resource_logger.get(
                    self.user_id, self.resource_id
                )
            else:
                old_response_data = self.resource_logger.get(
                    self.user_id, self.resource_id, iter
                )
            if old_response_data is None:
                return None
            _, response_model = pickle.loads(old_response_data)
            for element in response_model["process_log"]:
                self.module_output_dict[element["id"]] = element

            pc_step += response_model["progress"]["step"] - 1
        old_process_chain = response_model["process_chain_list"][0]

        return pc_step, old_process_chain

    def _validate_process_chain(
        self,
        process_chain=None,
        skip_permission_check=False,
        old_process_chain=None,
        pc_step=None,
    ):
        """
        Create the process list and check for user permissions.

        The following permissions are checked:

            - If the required modules are in the users module white list
            - If the user is allowed to access the required mapsets that
              are specified in the module input parameter
            - If the user is allowed to execute the number of processes

        Args:
            process_chain (dict): The process chain to be checked and converted
                                  into a process list
            skip_permission_check (bool): If set True, the permission checks
                                          of module access and process num
                                          limits are not performed
            old_process_chain (dict): The process chain of the previous
                                      resource run to be checked and converted
                                      for e.g. stdout

        Raises:
            This function raises AsyncProcessError in case of an error.

        Returns: list:
            The process list
        """

        if old_process_chain is not None:
            self.proc_chain_converter.process_chain_to_process_list(
                old_process_chain
            )
            self.proc_chain_converter.import_descr_list = list()
            self.proc_chain_converter.resource_export_list = list()

        # Backward compatibility
        if process_chain is None:
            process_list = (
                self.proc_chain_converter.process_chain_to_process_list(
                    self.request_data
                )
            )
            self.process_chain_list.append(self.request_data)
        else:
            process_list = (
                self.proc_chain_converter.process_chain_to_process_list(
                    process_chain
                )
            )
            self.process_chain_list.append(process_chain)
        if pc_step is not None:
            del process_list[:pc_step]

        # Check for the webhook
        if (
            hasattr(self.proc_chain_converter, "webhook_finished")
            and self.proc_chain_converter.webhook_finished is not None
        ):
            self.webhook_finished = self.proc_chain_converter.webhook_finished
        if (
            hasattr(self.proc_chain_converter, "webhook_update")
            and self.proc_chain_converter.webhook_update is not None
        ):
            self.webhook_update = self.proc_chain_converter.webhook_update
        if (
            hasattr(self.proc_chain_converter, "webhook_auth")
            and self.proc_chain_converter.webhook_auth is not None
        ):
            self.webhook_auth = self.proc_chain_converter.webhook_auth

        # Check for empty process chain
        if len(process_list) == 0 and len(self.resource_export_list) == 0:
            raise AsyncProcessError("Empty process chain, nothing to compute")

        # Check if the user is allowed to execute this number of processes
        if (
            skip_permission_check is False
            and len(process_list) > self.process_num_limit
        ):
            raise AsyncProcessError(
                "Process limit exceeded, a maximum of %i "
                "processes are allowed in the process chain."
                % self.process_num_limit
            )

        # Check if the module description was correct and if the
        # module or executable is in the user white list.
        for process in process_list:
            # Add the process to the internal list and dict
            # to access it in the python udf environment
            self._add_actinia_process(process)

            if process.exec_type == "grass" or process.exec_type == "exec":
                if skip_permission_check is False:
                    if process.skip_permission_check is False:
                        resp = check_location_mapset_module_access(
                            user_credentials=self.user_credentials,
                            config=self.config,
                            module_name=process.executable,
                        )
                        if resp is not None:
                            raise AsyncProcessError(
                                "Module or executable <%s> is not supported"
                                % process.executable
                            )
            else:
                message = (
                    "Wrong process description, type: %s "
                    "module/executable: %s, args: %s"
                    % (
                        str(process.exec_type),
                        str(process.executable),
                        str(process.executable_params),
                    )
                )
                raise AsyncProcessError(message)

        # Update the processing
        self._update_num_of_steps(len(process_list))

        return process_list

    def _setup(self, init_grass=True):
        """Setup the logger, the mapset lock and the credentials. Create the
        temporary grass database and temporary file directories

        ATTENTION: This method must be called first before any processing can
        take place

        What is done:

        - Create the resource and message logger
        - Create the redis lock interface for resource locking
        - Set cell limit, process number limit and process time limit from user
          credentials.
        - Create all required paths to original and temporary location and
          mapsets.
            - temp_location_path
            - global_location_path
            - grass_user_data_base <- This path will be created if it does not
              exist
            - user_location_path <- This path will be created if it does not
              exist
            - temp_grass_data_base <- This path will be created
            - temp_file_path <- This path will be created
            - Check if the current working location is in a persistent (global)
              GRASS GIS database (is_global_database)
        - Create the process chain to process list converter

        Args:
            init_grass (bool): Set true to initialize the user credentials
                               and the temporary database and location paths

        """
        # The setup should only be executed once
        if self.setup_flag is True:
            return
        else:
            self.setup_flag = True

        # fluent sender for this subprocess
        fluent_sender = None
        if self.has_fluent is True:
            from fluent import sender

            fluent_sender = sender.FluentSender(
                "actinia_core_logger",
                host=self.config.LOG_FLUENT_HOST,
                port=self.config.LOG_FLUENT_PORT,
            )
        kwargs = dict()
        kwargs["host"] = self.config.REDIS_SERVER_URL
        kwargs["port"] = self.config.REDIS_SERVER_PORT
        if (
            self.config.REDIS_SERVER_PW
            and self.config.REDIS_SERVER_PW is not None
        ):
            kwargs["password"] = self.config.REDIS_SERVER_PW
        self.resource_logger = ResourceLogger(
            **kwargs, fluent_sender=fluent_sender
        )

        self.message_logger = MessageLogger(
            config=self.config,
            user_id=self.user_id,
            fluent_sender=fluent_sender,
        )

        self.lock_interface = RedisLockingInterface()
        self.lock_interface.connect(**kwargs)
        del kwargs
        self.process_time_limit = int(
            self.user_credentials["permissions"]["process_time_limit"]
        )

        # Check and create all required paths to global, user and temporary
        # locations
        if init_grass is True:
            self._setup_paths()

        self.proc_chain_converter = ProcessChainConverter(
            config=self.config,
            temp_file_path=self.temp_file_path,
            process_dict=self.process_dict,
            temporary_pc_files=self.temporary_pc_files,
            required_mapsets=self.required_mapsets,
            resource_export_list=self.resource_export_list,
            output_parser_list=self.output_parser_list,
            message_logger=self.message_logger,
            send_resource_update=self._send_resource_update,
        )

    def _setup_paths(self):
        """Helper method to setup the paths"""
        self.cell_limit = int(
            self.user_credentials["permissions"]["cell_limit"]
        )
        self.process_num_limit = int(
            self.user_credentials["permissions"]["process_num_limit"]
        )
        # Setup the required paths
        self.temp_grass_data_base = os.path.join(
            self.grass_temp_database, self.temp_grass_data_base_name
        )
        self.temp_file_path = os.path.join(self.temp_grass_data_base, ".tmp")

        if self.location_name:
            self.temp_location_path = os.path.join(
                self.temp_grass_data_base, self.location_name
            )
            self.global_location_path = os.path.join(
                self.grass_data_base, self.location_name
            )
            # Create the user database path if it does not exist
            if not os.path.exists(self.grass_user_data_base):
                os.mkdir(self.grass_user_data_base)
            # Create the user group specific path, if it does not exist and set
            # the grass user database path accordingly
            self.grass_user_data_base = os.path.join(
                self.grass_user_data_base, self.user_group
            )
            if not os.path.exists(self.grass_user_data_base):
                os.mkdir(self.grass_user_data_base)
            # Create the user group specific location path, if it does not
            # exist
            self.user_location_path = os.path.join(
                self.grass_user_data_base, self.location_name
            )
            if not os.path.exists(self.user_location_path):
                os.mkdir(self.user_location_path)
            # Check if the location is located in the global database
            self.is_global_database = False
            location = os.path.join(self.grass_data_base, self.location_name)
            if os.path.isdir(location):
                self.is_global_database = True
        # Create the database, location and temporary file directories
        os.mkdir(self.temp_grass_data_base)
        os.mkdir(self.temp_file_path)

    def _create_temp_database(self, mapsets=None):
        """Create a temporary gis database with location and mapsets
        from the global and user group database for processing.

        IMPORTANT: All processing and mapaste management is performed within a
        temporary database!

        Link the required existing mapsets of global and user group locations
        into the temporary location directory.

        Linking is performed in two steps:
            1.) If the location is a global location, then the mapsets from the
                global location are linked in the temporary locations
            2.) Then link all required mapsets from the user group location
                into the temporary location

        Only mapsets from the global location are linked into the temporary
        location to which the user group has access.
        It checks for access in the global database but not in the user group
        database. The user can always access its own data of its group.

        Args:
            mapsets: A list of mapset names that should be linked into
                     the temporary location. If the list is empty, all
                     available user accessible mapsets of the global
                     and user group specific location will be linked.

        Raises:
            This function raises AsyncProcessError in case of an error.

        """
        # Assign default mapsets
        if mapsets is None:
            mapsets = []

        try:
            # Create the temporary location directory
            os.mkdir(self.temp_location_path)

            # Always link the PERMANENT mapset
            if len(mapsets) > 0 and "PERMANENT" not in mapsets:
                mapsets.append("PERMANENT")

            mapsets_to_link = []
            check_all_mapsets = False
            if not mapsets:
                check_all_mapsets = True

            # User and global location mapset linking
            self._link_mapsets(mapsets, mapsets_to_link, check_all_mapsets)

            # Check if we missed some of the required mapsets
            if check_all_mapsets is False:
                mapset_list = []
                for mapset_path, mapset in mapsets_to_link:
                    mapset_list.append(mapset)

                for mapset in mapsets:
                    if mapset not in mapset_list:
                        raise AsyncProcessError(
                            "Unable to link all required mapsets into "
                            "temporary location. Missing or un-accessible "
                            f"mapset <{mapset}> in location "
                            f"<{self.location_name}>"
                        )

            # Link the original mapsets from global and user database into the
            # temporary location
            for mapset_path, mapset in mapsets_to_link:
                if (
                    os.path.isdir(
                        os.path.join(self.temp_location_path, mapset)
                    )
                    is False
                ):
                    os.symlink(
                        mapset_path,
                        os.path.join(self.temp_location_path, mapset),
                    )

        except Exception as e:
            raise AsyncProcessError(
                "Unable to create a temporary GIS database"
                ", Exception: %s" % str(e)
            )

    def _link_mapsets(self, mapsets, mapsets_to_link, check_all_mapsets):
        """Helper method to link locations mapsets

        Args:
            mapsets (list): List of mapsets in location
            mapsets_to_link (list): List of mapsets paths to link
            check_all_mapsets (bool): If set True, the mapsets list is created
                                      with all locations on location_path

        Returns:
            mapsets (list): List of mapsets in location
            mapsets_to_link (list): List of mapsets paths to link
        """
        # Global location mapset linking
        if self.is_global_database is True:
            # List all available mapsets in the global location
            mapsets, mapsets_to_link = self._list_all_available_mapsets(
                self.global_location_path,
                mapsets,
                check_all_mapsets,
                mapsets_to_link,
                True,
            )
        # Check for leftover mapsets
        left_over_mapsets = []
        for mapset in mapsets:
            if mapset not in mapsets_to_link:
                left_over_mapsets.append(mapset)
        # List all available mapsets in the user location
        mapsets, mapsets_to_link = self._list_all_available_mapsets(
            self.user_location_path,
            left_over_mapsets,
            check_all_mapsets,
            mapsets_to_link,
            False,
        )
        return mapsets, mapsets_to_link

    def _list_all_available_mapsets(
        self,
        location_path,
        mapsets,
        check_all_mapsets,
        mapsets_to_link,
        global_db=False,
    ):
        """Helper method to list all available mapsets and for global database
        it is checked if the mapset can be accessed.

        Args:
            location_path (str): Path to location (global or user)
            mapsets (list): List of mapsets names to link.
                            The mapsets list can be empty, if check_all_mapsets
                            is True the list is filled with all mapsets from
                            the location_path
            check_all_mapsets (bool): If set True, the mapsets list is created
                                      with all locations on location_path
            mapsets_to_link (list): List of mapset paths to link
            global_db (bool): If set True, the location/mapset access is
                              checked

        Returns:
            mapsets (list): List of mapsets in location
            mapsets_to_link (list): List of mapsets paths to link
        """
        if os.path.isdir(location_path):
            if check_all_mapsets is True:
                mapsets = os.listdir(location_path)
            for mapset in mapsets:
                mapset_path = os.path.join(location_path, mapset)
                if os.path.isdir(mapset_path) and os.access(
                    mapset_path, os.R_OK & os.X_OK
                ):
                    # Check if a WIND file exists to be sure it is a mapset
                    if (
                        os.path.isfile(os.path.join(mapset_path, "WIND"))
                        is True
                    ):
                        if mapset not in mapsets_to_link and global_db is True:
                            # Link the mapset from the global database
                            # only if it can be accessed
                            resp = check_location_mapset_module_access(
                                user_credentials=self.user_credentials,
                                config=self.config,
                                location_name=self.location_name,
                                mapset_name=mapset,
                            )
                            if resp is None:
                                mapsets_to_link.append((mapset_path, mapset))
                        elif (
                            mapset not in mapsets_to_link
                            and global_db is False
                        ):
                            mapsets_to_link.append((mapset_path, mapset))
                    else:
                        raise AsyncProcessError(
                            "Invalid mapset <%s> in location <%s>"
                            % (mapset, self.location_name)
                        )
        else:
            if global_db is True:
                msg = (
                    "Unable to access global location <%s>"
                    % self.location_name
                )
            else:
                msg = (
                    "Unable to access user location <%s>" % self.location_name
                )
            raise AsyncProcessError(msg)
        return mapsets, mapsets_to_link

    def _create_grass_environment(
        self, grass_data_base, mapset_name="PERMANENT"
    ):
        """Sets up the GRASS environment to run modules

        Args:
            mapset_name (str): The mapset that should used for processing,
                               default is PERMANENT

        Raises:
            This method will raise an AsyncProcessError if the mapset is
            not in the access list of the user.

        """
        self.message_logger.info(
            "Initlialize GRASS grass_data_base: %s; location: %s; mapset: %s"
            % (grass_data_base, self.location_name, mapset_name)
        )

        self.ginit = GrassInitializer(
            grass_data_base=grass_data_base,
            grass_base_dir=self.config.GRASS_GIS_BASE,
            location_name=self.location_name,
            mapset_name=mapset_name,
            config=self.config,
            grass_addon_path=self.config.GRASS_ADDON_PATH,
            user_id=self.user_id,
        )

        self.ginit.initialize()

    def _create_temporary_mapset(
        self,
        temp_mapset_name,
        source_mapset_name=None,
        interim_result_mapset=None,
        interim_result_file_path=None,
    ):
        """Create the temporary mapset and switch into it

        This method needs an initialized the GRASS environment.

        It will check access to all required mapsets and adds them to the
        mapset search path.

        IMPORTANT: You need to call self._create_grass_environment() to set up
        the environment before calling this method.

        A new temporary mapset is created. All in the process chain detected
        mapsets of input maps or STDS will be added to the mapset search path
        using g.mapsets.

        Optionally the WIND file of a source mapset can be copied into the
        temporary mapset.

        Args:
            temp_mapset_name (str): The name of the temporary mapset to be
                                    created
            source_mapset_name (str): The name of the source mapset to copy the
                                      WIND file from
            interim_result_mapset (str): The path to the mapset which is saved
                                         as interim result and should be used
                                         as start mapset for the job resumtion
            interim_result_file_path (str): The path of the interim result
                                            temporary file path


        Raises:
            This function will raise an exception if the
            g.mapset/g.mapsets/db.connect modules fail

        """
        self.temp_mapset_path = os.path.join(
            self.temp_location_path, temp_mapset_name
        )

        # if interim_result_mapset is set copy the mapset from the interim
        # results
        if interim_result_mapset:
            self.message_logger.info(
                "Rsync interim result mapset to temporary GRASS DB"
            )
            # change mapset name for groups, raster VRTs and tgis
            for directory in ["group", "cell_misc", "tgis"]:
                change_mapsetname(
                    os.path.join(interim_result_mapset, directory),
                    directory,
                    os.path.basename(interim_result_mapset),
                    os.path.basename(self.temp_mapset_path),
                )
            rsync_status = self.interim_result.rsync_mapsets(
                interim_result_mapset, self.temp_mapset_path
            )
            if rsync_status != "success":
                raise RsyncError(
                    "Error while rsyncing of interim results to new temporare "
                    "mapset"
                )
            self.interim_result.rsync_additional_mapsets(
                os.path.dirname(self.temp_mapset_path)
            )
        if interim_result_file_path:
            self.message_logger.info(
                "Rsync interim result file path to temporary GRASS DB"
            )
            rsync_status = self.interim_result.rsync_mapsets(
                interim_result_file_path, self.temp_file_path
            )
            if rsync_status != "success":
                raise RsyncError(
                    "Error while rsyncing of interim temporary file path to "
                    "new temporare file path"
                )

        self.ginit.run_module(
            "g.mapset", ["-c", "mapset=%s" % temp_mapset_name]
        )

        if self.required_mapsets:
            self.ginit.run_module(
                "g.mapsets",
                [
                    "operation=add",
                    "mapset=%s" % (",".join(self.required_mapsets)),
                ],
            )

            self.message_logger.info(
                "Added the following mapsets to the mapset "
                "search path: " + ",".join(self.required_mapsets)
            )

        # Set the vector database connection to vector map specific databases
        self.ginit.run_module(
            "db.connect",
            [
                "driver=sqlite",
                "database=$GISDBASE/$LOCATION_NAME/$MAPSET/vector/$MAP/"
                "sqlite.db",
            ],
        )

        # self.ginit.run_module("g.gisenv", ["set=DEBUG=2",])

        # If a source mapset is provided, the WIND file will be copied from it
        # to the temporary mapset
        if source_mapset_name is not None and interim_result_mapset is None:
            source_mapset_path = os.path.join(
                self.temp_location_path, source_mapset_name
            )
            if os.path.exists(os.path.join(source_mapset_path, "WIND")):
                shutil.copyfile(
                    os.path.join(source_mapset_path, "WIND"),
                    os.path.join(self.temp_mapset_path, "WIND"),
                )

    def _cleanup(self):
        """Clean up the GrassInitializer files created in
        self._setup() and remove the created temporary database.

        """
        if self.ginit:
            self.ginit.clean_up()

        if (
            self.temp_grass_data_base is not None
            and os.path.exists(self.temp_grass_data_base)
            and os.path.isdir(self.temp_grass_data_base)
        ):
            shutil.rmtree(self.temp_grass_data_base, ignore_errors=True)

    def _check_reset_region(self):
        """Check the current region settings against the user cell limit.

        Reset the current processing region to a meaningful state
        so that the user cell limit is not reached and the mapset can be
        accessed again.

        Raises:
            This method will raise an AsyncProcessError exception

        """
        if self.skip_region_check is True:
            return

        errorid, stdout_buff, stderr_buff = self.ginit.run_module(
            "g.region", ["-ug"]
        )

        if errorid != 0:
            raise AsyncProcessError(
                "Unable to check the computational region size"
            )

        str_list = stdout_buff.split()
        region = {}
        for line in str_list:
            if "=" in line:
                option = line.split("=", 1)
                region[option[0]] = option[1]

        self.message_logger.info(str(region))

        num_cells = int(region["cells"])
        ns_res = float(region["nsres"])
        ew_res = float(region["ewres"])

        if num_cells > self.cell_limit:
            self._adjust_region_size(num_cells, ns_res, ew_res)

    def _adjust_region_size(self, num_cells, ns_res, ew_res):
        """Helper method to adjust the region size

        Args:
            num_cells (int): GRASS GIS number of cells of the region
            ns_res (float): GRASS GIS north-south cell resolution of the region
            ew_res (float): GRASS GIS east-west cell resolution of the region

        Raises:
            This method will raise an AsyncProcessError exception

        """
        fak = num_cells / self.cell_limit
        fak += 2.0
        fak = math.sqrt(fak) + 2.0
        ns_res = ns_res * fak
        ew_res = ew_res * fak
        errorid, stdout_buff, stderr_buff = self.ginit.run_module(
            "g.region", ["nsres=%f" % ns_res, "ewres=%f" % ew_res, "-g"]
        )
        self.message_logger.info(stdout_buff)
        if errorid != 0:
            raise AsyncProcessError(
                "Unable to adjust the region settings to nsres: "
                "%f ewres: %f error: %s" % (ns_res, ew_res, stderr_buff)
            )
        raise AsyncProcessError(
            "Region too large, set a coarser resolution to minimum nsres: "
            "%f ewres: %f [num_cells: %d]" % (ns_res, ew_res, num_cells)
        )

    def _increment_progress(self, num=1):
        """Increment the progress step by a specific number

        Args:
            num (int): The number for which the progress should be increased
        """
        self.progress_steps += num
        self.progress["step"] = self.progress_steps

    def _add_actinia_process(self, process: Process):
        """Add an actinia process to the list and dictionary

        Args:
            process: The actinia process
        """
        self.actinia_process_dict[process.id] = process
        self.actinia_process_list.append(process)

    def _update_num_of_steps(self, num):
        """Update the number of total steps

        Args:
            num: The number of processes to be added to the total number of
                 processes
        """
        self.number_of_processes += num
        self.progress["num_of_steps"] = self.number_of_processes

    def _wait_for_process(
        self, module_name, module_parameter, proc, poll_time
    ):
        """
        Wait for a specific process. Catch termination requests, process
        time limits and send updates to the user.

        Args:
            module_name: The name of the GRASS module or executable
            module_parameter: The parameter of a GRASS module or a executable
            proc: The process to wait for and monitor
            poll_time: The poll time to send updates and check for termination

        Returns:
            (float)
            The run time in seconds

        """

        start_time = time.time()

        termination_check_count = 0
        update_check_count = 0
        while True:
            if proc.poll() is not None:
                break
            else:
                # Sleep some time and update the resource status
                time.sleep(poll_time)
                termination_check_count += 1
                update_check_count += 1

                # Check all 10 loops for termination
                if termination_check_count == 10:
                    termination_check_count = 0
                    # check if the resource should be terminated
                    # and kill the current process
                    if (
                        self.resource_logger.get_termination(
                            self.user_id, self.resource_id, self.iteration
                        )
                        is True
                    ):
                        proc.kill()
                        raise AsyncProcessTermination(
                            "Process <%s> was terminated "
                            "by user request" % module_name
                        )

                # Send all 100 loops a status update
                if update_check_count == 100:
                    update_check_count = 0
                    # Check max runtime of process
                    curr_time = time.time()
                    if (curr_time - start_time) > self.process_time_limit:
                        proc.kill()
                        raise AsyncProcessTimeLimit(
                            "Time (%i seconds) exceeded to run executable %s"
                            % (self.process_time_limit, module_name)
                        )
                    # Reduce the length of the command line parameters for
                    # lesser logging overhead
                    mparams = str(module_parameter)
                    if len(mparams) > 100:
                        mparams = "%s ... %s" % (mparams[0:50], mparams[-50:])
                    message = (
                        f"Running executable {module_name} with parameters "
                        f"{mparams} for {curr_time - start_time} seconds"
                    )
                    self._send_resource_update(message)

        return time.time() - start_time

    def _run_process(self, process, poll_time=0.05):
        """
        Run a process actinia_core.core.common.process_object.Process) with
        options and send progress updates to the resource database.

        IMPORTANT: Use this method to run programs that are not GRASS modules.

        Check each poll the termination status of the resource.
        If the termination state is set True, terminate the current process
        and raise an AsyncProcessTermination exception that must be caught
        by the run() method.

        Args:
            process actinia_core.core.common.process_object.Process):
                The process object that should be executed
            poll_time (float): The time to check the process status and to send
                               updates to the resource db

        Raises:
            AsyncProcessError:
            AsyncProcessTermination:
            AsyncProcessTimeLimit:

        Returns:
            tuple:
            (returncode, stdout_buff, stderr_buff)

        """
        if (
            self.resource_logger.get_termination(
                self.user_id, self.resource_id, self.iteration
            )
            is True
        ):
            raise AsyncProcessTermination(
                "Process <%s> was terminated by "
                "user request" % process.executable
            )

        return self._run_executable(process, poll_time)

    def _run_module(self, process, poll_time=0.05):
        """Run the GRASS module actinia_core.core.common.process_object.Process
        with its module options and send progress updates to the database
        server that manages the resource entries.

        Check before each module run the size of the region. If the maximum
        number of cells are exceeded then raise an AsyncProcessError exception
        that the maximum number of cells are exceeded and reset them to a
        meaningful state, so that the user can still run processes in the
        mapset.

        The region is checked for the first module of a process chain and after
        that, for each g.region call that was present in the process chain.

        Check each poll the termination status of the resource.
        If the termination state is set True, terminate the current process
        and raise an AsyncProcessTermination exception that must be caught
        by the run() method.

        By default the status of the running process is checked each 0.005
        seconds. This is the minimum amount of time a process can run. If many
        tiny running processes are executed in a large process chain, then this
        value must be adjusted.

        Args:
            process actinia_core.core.common.process_object.Process):
                The process object that should be executed
            poll_time (float): The time to check the process status and to send
                               updates to the resource db

        Raises:
            AsyncProcessError:
            AsyncProcessTermination:
            AsyncProcessTimeLimit:

        Returns:
            tuple:
            (returncode, stdout_buff, stderr_buff)

        """
        # Count the processes
        self.process_count += 1
        # Check for each 20. process if a kill request was received
        # This is required in case a single of many fast running processes in a
        # chain is not able to trigger the termination check in the while loop
        if self.process_count % 20 == 0:
            if (
                self.resource_logger.get_termination(
                    self.user_id, self.resource_id, self.iteration
                )
                is True
            ):
                raise AsyncProcessTermination(
                    "Process <%s> was terminated "
                    "by user request" % process.executable
                )

            message = "Running module %s with parameters %s" % (
                process.executable,
                str(process.executable_params),
            )
            self._send_resource_update(message)

        # Check reset region if a g.region call was present in the process
        # chain. By default the initial value of last_module is "g.region" to
        # assure for first run of a process from the process chain, the
        # region settings are evaluated
        if (
            self.last_module == "g.region"
            and process.skip_permission_check is False
        ):
            self._check_reset_region()
        # Save the last module name. This is needed to check the region
        # settings
        self.last_module = process.executable

        return self._run_executable(process, poll_time)

    def _run_executable(self, process, poll_time=0.005):
        """Runs a GRASS module or aactinia_core.core.Unix executable and sets
        up the correct handling of stdout, stderr and stdin, creates the
        process log model and returns stdout, stderr and the return code.

        It creates the temporary file paths.

        The returncode of 0 indicates that it ran successfully. A negative
        value -N indicates that the child was terminated by signal N (POSIX
        only; see also
        https://en.wikipedia.org/wiki/Signal_(IPC)#Default_action).

        Args:
            process actinia_core.core.common.process_object.Process):
                The process object that should be executed
            poll_time (float): The time to check the process status and to send
                               updates to the resource db

        Raises:
            AsyncProcessError:
            AsyncProcessTermination:
            AsyncProcessTimeLimit:

        Returns:
            tuple:
            (returncode, stdout_buff, stderr_buff)

        """

        # Use temporary files to catch stdout and stderr
        stdout_buff = tempfile.NamedTemporaryFile(
            mode="w+b", delete=True, dir=self.temp_file_path
        )
        stderr_buff = tempfile.NamedTemporaryFile(
            mode="w+b", delete=True, dir=self.temp_file_path
        )
        stdin_file = None

        if process.stdin_source is not None:
            tmp_file = self.proc_chain_converter.generate_temp_file_path()
            stdin_file = open(tmp_file, "w")
            stdin_file.write(process.stdin_source())
            stdin_file.close()
            stdin_file = open(tmp_file, "r")

        self._increment_progress(num=1)

        # GRASS andactinia_core.core.Unix executables have different run
        # methods
        if process.exec_type in "grass":
            proc = self.ginit.run_module(
                process.executable,
                process.executable_params,
                raw=True,
                stdout=stdout_buff,
                stderr=stderr_buff,
                stdin=stdin_file,
            )
        else:
            inputlist = list()
            inputlist.append(process.executable)
            inputlist.extend(process.executable_params)

            proc = subprocess.Popen(
                args=inputlist,
                stdout=stdout_buff,
                stderr=stderr_buff,
                stdin=stdin_file,
            )

        run_time = self._wait_for_process(
            process.executable, process.executable_params, proc, poll_time
        )

        proc.wait()

        stdout_buff.flush()
        stderr_buff.flush()
        stdout_buff.seek(0)
        stderr_buff.seek(0)
        stdout_string = stdout_buff.read().decode()
        stderr_string = stderr_buff.read().decode()
        stdout_buff.close()
        stderr_buff.close()
        if stdin_file:
            stdin_file.close()

        process.set_stdouts(stdout=stdout_string, stderr=stderr_string)

        kwargs = {
            "id": process.id,
            "executable": process.executable,
            "parameter": process.executable_params,
            "return_code": proc.returncode,
            "stdout": stdout_string,
            "stderr": stderr_string.split("\n"),
            "run_time": run_time,
        }
        if self.temp_mapset_path:
            kwargs["mapset_size"] = get_directory_size(self.temp_mapset_path)

        plm = ProcessLogModel(**kwargs)

        self.module_output_log.append(plm)
        # Store the log in an additional dictionary for automated output
        # generation
        if process.id is not None:
            self.module_output_dict[process.id] = plm

        if proc.returncode != 0:
            raise AsyncProcessError(
                "Error while running executable <%s>" % process.executable
            )

        # save interim results
        if (
            self.interim_result.saving_interim_results is True
            and self.temp_mapset_path is not None
        ):
            self.interim_result.save_interim_results(
                self.progress_steps, self.temp_mapset_path, self.temp_file_path
            )
        elif self.temp_mapset_path is None:
            self.message_logger.debug(
                "No temp mapset path set. Because of that no interim results"
                " can be saved!"
            )

        return proc.returncode, stdout_string, stderr_string

    def _create_temporary_grass_environment(
        self,
        source_mapset_name=None,
        interim_result_mapset=None,
        interim_result_file_path=None,
    ):
        """Create a temporary GRASS GIS environment

        This method will:
            1. create the temporary database
            2. sets-up the GRASS environment
            3. Create temporary mapset

        This method will link the required mapsets that are
        defined in *self.required_mapsets* into the location.
        The mapsets may be from the global and/or user database.

        Args:
            source_mapset_name (str): The name of the source mapset to copy the
                                      WIND file from
            interim_result_mapset (str): The path to the mapset which is saved
                                         as interim result and should be used
                                         as start mapset for the job resumtion
            interim_result_file_path (str): The path of the interim result
                                            temporary file path
        Raises:
            This method will raise an AsyncProcessError
        """

        # Create the temp database and link the
        # required mapsets into it
        self._create_temp_database(self.required_mapsets)

        # Initialize the GRASS environment and switch into PERMANENT
        # mapset, which is always linked
        self._create_grass_environment(
            grass_data_base=self.temp_grass_data_base, mapset_name="PERMANENT"
        )

        # Create the temporary mapset and switch into it
        self._create_temporary_mapset(
            temp_mapset_name=self.temp_mapset_name,
            source_mapset_name=source_mapset_name,
            interim_result_mapset=interim_result_mapset,
            interim_result_file_path=interim_result_file_path,
        )

    def _execute(self, skip_permission_check=False):
        """Overwrite this function in subclasses.

        This function will be executed by the run() function

        - Setup logger and credentials
        - Analyse the process chain
        - Create the temporal database
        - Initialize the GRASS environment and create the temporary mapset
        - Run the modules
        - Parse the stdout output of the modules and generate the module
          results

        Args:
            skip_permission_check (bool): If set True, the permission checks of
                                          module access and process num
                                          limits are not performed

        Raises:
            This method will raise an AsyncProcessError, AsyncProcessTimeLimit
            or AsyncProcessTermination

        """
        # Create the process chain
        if self.rdc.iteration is not None:
            process_list = self._create_temporary_grass_environment_and_process_list_for_iteration(
                skip_permission_check=skip_permission_check
            )
        else:
            process_list = (
                self._create_temporary_grass_environment_and_process_list(
                    skip_permission_check=skip_permission_check
                )
            )

        # Run all executables
        self._execute_process_list(process_list=process_list)
        # Parse the module sdtout outputs and create the results
        self._parse_module_outputs()

    def _create_temporary_grass_environment_and_process_list_for_iteration(
        self, process_chain=None, skip_permission_check=False
    ):
        """Helper method to:

        - Setup logger and credentials
        - Analyse the process chain
        - Create the temporal database
        - Initialize the GRASS environment and create the temporary mapset
        - Return the created process list

        Args:
            process_chain (dict): The process chain to be checked and converted
                                  into a process list
            skip_permission_check (bool): If set True, the permission checks
                                          of module access and process num
                                          limits are not performed

        Raises:
            This method will raise an AsyncProcessError

        Returns: list
                  The process list to be executed by _execute_process_list()

        """
        # Setup the user credentials and logger
        self._setup()

        # Create and check the process chain
        (
            pc_step,
            old_process_chain_list,
        ) = self._get_previous_iteration_process_chain()
        self.interim_result.set_old_pc_step(pc_step)
        process_list = self._validate_process_chain(
            process_chain=self.request_data,
            old_process_chain=old_process_chain_list,
            skip_permission_check=skip_permission_check,
            pc_step=pc_step,
        )

        # check iterim results
        (
            interim_result_mapset,
            interim_result_file_path,
        ) = self.interim_result.check_interim_result_mapset(
            pc_step, self.iteration - 1
        )
        if interim_result_mapset is None:
            return None

        # Init GRASS and create the temporary mapset with the interim results
        self._create_temporary_grass_environment(
            interim_result_mapset=interim_result_mapset,
            interim_result_file_path=interim_result_file_path,
        )

        return process_list

    def _create_temporary_grass_environment_and_process_list(
        self, process_chain=None, skip_permission_check=False
    ):
        """Helper method to:

        - Setup logger and credentials
        - Analyse the process chain
        - Create the temporal database
        - Initialize the GRASS environment and create the temporary mapset
        - Return the created process list

        Args:
            process_chain (dict): The process chain to be checked and converted
                                  into a process list
            skip_permission_check (bool): If set True, the permission checks
                                          of module access and process num
                                          limits are not performed

        Raises:
            This method will raise an AsyncProcessError

        Returns: list
                  The process list to be executed by _execute_process_list()

        """
        # Setup the user credentials and logger
        self._setup()

        # Create and check the process chain
        process_list = self._validate_process_chain(
            process_chain=process_chain,
            skip_permission_check=skip_permission_check,
        )

        # Init GRASS and create the temporary mapset
        self._create_temporary_grass_environment()

        return process_list

    def _parse_module_outputs(self):
        """Parse the module stdout outputs and parse them into the required
        formats: table, list or kv

        This functions analyzes the output_parser_list for entries to parse.
        It will convert the stdout strings into tables, lists or key/value
        outputs and stores the result in the module_result dictionary using the
        provided id of the StdoutParser.

        """

        for entry in self.output_parser_list:
            for process_id, stdout_def in entry.items():
                id = stdout_def["id"]
                format = stdout_def["format"]
                delimiter = stdout_def["delimiter"]
                if process_id not in self.module_output_dict:
                    raise AsyncProcessError(
                        "Unable to find process id in module output dictionary"
                    )
                stdout = self.module_output_dict[process_id]["stdout"]
                # Split the rows by the \n new line delimiter
                rows = stdout.strip().split("\n")
                if "table" in format:
                    result = []
                    for row in rows:
                        row = row.strip()
                        values = row.split(delimiter)
                        value_list = []
                        for value in values:
                            value_list.append(value.strip())
                        result.append(value_list)
                elif "list" in format:
                    result = []
                    for row in rows:
                        value = row.strip()
                        result.append(value)
                elif "kv" in format:
                    result = dict()
                    for row in rows:
                        row = row.strip()
                        key, value = row.split(delimiter, 1)
                        result[key.strip()] = value.strip()
                elif "json" in format:
                    result = None
                    try:
                        result = {
                            i[0]: i[1]
                            for i in [
                                entry.split(delimiter, 1)
                                for entry in stdout.strip("\n").split("\n")
                            ]
                        }
                    except Exception:
                        try:
                            result = json.loads(stdout)
                        except Exception:
                            pass
                    finally:
                        if not result:
                            result = stdout
                else:
                    raise AsyncProcessError("Wrong stdout parser format")

                # Store the parser result
                self.module_results[id] = result

    def _execute_process_list(self, process_list):
        """
        Run all modules or executables that are specified in the process list

        Args:
            process_list: The process list that was generated by
                          _validate_process_chain() which is also called in
                          _create_temporary_grass_environment_and_process_list()

        Raises:
            This method will raise an AsyncProcessError, AsyncProcessTimeLimit
            or AsyncProcessTermination

        """
        for process in process_list:
            if process.exec_type == "grass":
                self._run_module(process)
            elif process.exec_type == "exec":
                self._run_process(process)
            elif process.exec_type == "python":
                eval(process.executable)

    def _interim_results(self):
        """Check if interim results should be saved or cleaned up"""

        if (
            "error" in self.run_state
            and self.interim_result.saving_interim_results == "onError"
        ):
            if self.progress_steps > 0:
                self.interim_result.delete_interim_results()
                self.interim_result.save_interim_results(
                    self.progress_steps - 1,
                    self.temp_mapset_path,
                    self.temp_file_path,
                    force_copy=True,
                )
        elif (
            "success" in self.run_state
            and self.interim_result.saving_interim_results is not False
        ):
            self.interim_result.delete_interim_results()

    def _final_cleanup(self):
        """Overwrite this function in subclasses to perform the final cleanup,
        by default this function calls self._cleanup() to remove the temporary
        gis database.
        This function should not raise any exceptions. Extend the cleaning
        functionality here.
        """
        # Clean up and remove the temporary gisdbase
        self._cleanup()

    def run(self):
        """This function will run the processing and will catch and process
        any exceptions that were raised while processing. Call this function to
        run the processing.

        You have to implement/overwrite two methods that are called here:

            * self._execute()
            * self._interim_results()
            * self._final_cleanup()

            e_type, e_value, e_traceback = sys.exc_info()
            message = [e.__class__, e_type, e_value, traceback.format_tb(
                       e_traceback)]
            message = pprint.pformat(message)
        """

        try:
            # Run the _execute function that does all the work
            self._execute()
        except AsyncProcessTermination as e:
            self.run_state = {"terminated": str(e)}
        except AsyncProcessTimeLimit as e:
            self.run_state = {"time limit exceeded": str(e)}
        except AsyncProcessError as e:
            e_type, e_value, e_tb = sys.exc_info()
            model = ExceptionTracebackModel(
                message=str(e_value),
                traceback=traceback.format_tb(e_tb),
                type=str(e_type),
            )
            self.run_state = {"error": str(e), "exception": model}
        except KeyboardInterrupt as e:
            e_type, e_value, e_tb = sys.exc_info()
            model = ExceptionTracebackModel(
                message=str(e_value),
                traceback=traceback.format_tb(e_tb),
                type=str(e_type),
            )
            self.run_state = {"error": str(e), "exception": model}
        except Exception as e:
            e_type, e_value, e_tb = sys.exc_info()
            model = ExceptionTracebackModel(
                message=str(e_value),
                traceback=traceback.format_tb(e_tb),
                type=str(e_type),
            )
            self.run_state = {"error": str(e), "exception": model}
        finally:
            try:
                # Check if interim results should be saved and save them
                self._interim_results()
                # Call the final cleanup, before sending the status messages
                self._final_cleanup()
            except Exception as e:
                e_type, e_value, e_tb = sys.exc_info()
                model = ExceptionTracebackModel(
                    message=str(e_value),
                    traceback=traceback.format_tb(e_tb),
                    type=str(e_type),
                )
                self.run_state = {"error": str(e), "exception": model}
            # After all processing finished, send the final status
            if "success" in self.run_state:
                self._send_resource_finished(
                    message=self.finish_message, results=self.module_results
                )
            elif "terminated" in self.run_state:
                # Send an error message if an exception was raised
                self._send_resource_terminated(
                    message=self.run_state["terminated"]
                )
            elif "time limit exceeded" in self.run_state:
                self._send_resource_time_limit_exceeded(
                    message=self.run_state["time limit exceeded"]
                )
            elif "error" in self.run_state:
                # Send an error message if an exception was raised
                self._send_resource_error(
                    message=self.run_state["error"],
                    exception=self.run_state["exception"],
                )
            else:
                self._send_resource_error(message="Unknown error")



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Tests: Async process custom test case admin
"""
import unittest
from flask.json import dumps as json_dumps

try:
    from .test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX
except ModuleNotFoundError:
    from test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Sören Gebbert"
__email__ = "soerengebbert@googlemail.com"


class AsyncProcessCustomTestCaseAdmin(ActiniaResourceTestCaseBase):
    def test_async_processing(self):
        rv = self.server.post(
            URL_PREFIX + "/custom_process/uname",
            headers=self.admin_auth_header,
            data=json_dumps(["-a"]),
            content_type="application/json",
        )
        self.waitAsyncStatusAssertHTTP(rv, headers=self.admin_auth_header)

    def test_async_processing_error_1(self):
        """
        Test the error handling in case the program was not found
        Returns:

        """
        rv = self.server.post(
            URL_PREFIX + "/custom_process/uname_nopo",
            headers=self.admin_auth_header,
            data=json_dumps(["-a"]),
            content_type="application/json",
        )
        self.waitAsyncStatusAssertHTTP(
            rv, headers=self.admin_auth_header, http_status=400, status="error"
        )

    def test_async_processing_error_2(self):
        """
        Test the error handling in case the program was not found
        Returns:

        """
        rv = self.server.post(
            URL_PREFIX + "/custom_process/cat",
            headers=self.admin_auth_header,
            data=json_dumps(["non_existing_file"]),
            content_type="application/json",
        )
        self.waitAsyncStatusAssertHTTP(
            rv,
            headers=self.admin_auth_header,
            http_status=400,
            status="error",
            message_check="AsyncProcessError:",
        )

    def test_async_processing_user(self):
        """
        Sleep for one secnond
        Returns:

        """
        rv = self.server.post(
            URL_PREFIX + "/custom_process/sleep",
            headers=self.user_auth_header,
            data=json_dumps(["1"]),
            content_type="application/json",
        )
        self.waitAsyncStatusAssertHTTP(rv, headers=self.admin_auth_header)

    def test_async_processing_error_to_long(self):
        """
        Test the error handling in case the program is running too long
        Returns:

        """
        rv = self.server.post(
            URL_PREFIX + "/custom_process/sleep",
            headers=self.user_auth_header,
            data=json_dumps(["10"]),
            content_type="application/json",
        )
        self.waitAsyncStatusAssertHTTP(
            rv,
            headers=self.admin_auth_header,
            http_status=400,
            status="terminated",
            message_check="AsyncProcessTimeLimit:",
        )

    def test_async_processing_unauthorized_access(self):
        """
        Test the error handling in case the program is is not in access list
        Returns:

        """
        rv = self.server.post(
            URL_PREFIX + "/custom_process/ps",
            headers=self.user_auth_header,
            data=json_dumps(["-rf"]),
            content_type="application/json",
        )
        self.waitAsyncStatusAssertHTTP(
            rv,
            headers=self.admin_auth_header,
            http_status=400,
            status="error",
            message_check="AsyncProcessError:",
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Soeren Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Purpose: Command line program to manage actinia user
         entries in the Redis database
"""

from actinia_core.core.common.config import Configuration
from actinia_core.core.common.user import ActiniaUser
from actinia_core.core.redis_user import redis_user_interface
import argparse
import pprint
import sys

__license__ = "GPLv3"
__author__ = "Soeren Gebbert"
__copyright__ = (
    "Copyright 2016-2023, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


def verify_password(username_or_token, password):
    """Verify the user name and password.

    Instead of a user name an authentication token
    can be provided.

    Args:
        username_or_token (str): The username or an authentication token
        password (str): The optional user password, not required in case
        of token

    Returns:
        bool: True if authorized or False if not

    """
    # first try to authenticate by token
    user = ActiniaUser.verify_auth_token(username_or_token)
    if not user:
        # try to authenticate with username/password
        user = ActiniaUser(user_id=username_or_token)
        if not user.exists() or not user.verify_password(password):
            return False
    return True


def set_user_credentials(user, args, method):
    """Set the user credentials

    Args:
        user (actinia.core.user.ActiniaUser()): The user object
        args: The command line arguments
        method (str): "create", "update", "update_add",
                      "update_rm", "update_rm_location"

    """

    if args.role:
        user.set_role(args.role)
    if args.cell_limit:
        user.set_cell_limit(args.cell_limit)
    if args.process_num_limit:
        user.set_process_num_limit(args.process_num_limit)
    if args.process_time_limit:
        user.set_process_time_limit(args.process_time_limit)
    if args.password:
        user.hash_password(args.password)

    if args.datasets:
        datasets = {}
        lm_list = args.datasets.split(",")

        for lm in lm_list:
            if "/" not in lm:
                sys.stderr.write("Wrong location mapset format for datasets\n")
                return

            if "/" in lm:
                location, mapset = lm.split("/")
            else:
                location = lm
                mapset = None

            if location not in datasets:
                datasets[location] = []

            datasets[location].append(mapset)

        if method == "update" or method == "create":
            user.set_accessible_datasets(datasets)
        if method == "update_add":
            for dataset in datasets:
                user.add_accessible_dataset(dataset, datasets[dataset])
        if method == "update_rm":
            for dataset in datasets:
                user.remove_mapsets_from_location(dataset, datasets[dataset])
        if method == "update_rm_location":
            for dataset in datasets:
                user.remove_location(dataset)

    if args.modules:
        modules = args.modules.split(",")
        if method == "update" or method == "create":
            user.set_accessible_modules(modules)
        if method == "update_add":
            user.add_accessible_modules(modules)
        if method == "update_rm":
            user.remove_accessible_modules(modules)


def main():
    """User management"""
    parser = argparse.ArgumentParser(
        description="Manage actinia users in the Redis database.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "action",
        type=str,
        default="create",
        choices=[
            "create",
            "delete",
            "show",
            "list",
            "update",
            "update_add",
            "update_rm",
            "update_rm_location",
            "pwcheck",
        ],
        help="The action that should be performed:"
        " * create: Create a user"
        " * show:   Show the user credentials"
        " * list:   List all users"
        " * update: Update a user with the provided parameters"
        " * update_add: Update a user with the provided parameters"
        " and add datasets or/and modules"
        " * update_rm: Update a user with the provided parameters"
        " and remove mapsets or/and modules"
        " * update_rm_location: Update a user with the provided parameters"
        " and remove locations or/and modules"
        " * pwcheck: Check the password of the user",
    )
    parser.add_argument(
        "-s",
        "--server",
        type=str,
        required=False,
        help="The host name of the redis server,"
        " default is the value from the actinia config file",
    )
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        required=False,
        help="The port of the redis server,"
        " default is the value from the actinia config file",
    )
    parser.add_argument(
        "-a",
        "--redis_password",
        type=str,
        required=False,
        help="The password of the redis server,"
        " default is the value from the actinia config file or None",
    )
    parser.add_argument(
        "-u",
        "--user_id",
        type=str,
        required=False,
        help="The user name",
    )
    parser.add_argument(
        "-g",
        "--user_group",
        type=str,
        required=False,
        help="The name of the user group this user is associated with",
    )
    parser.add_argument(
        "-w",
        "--password",
        type=str,
        required=False,
        help="The password",
    )
    parser.add_argument(
        "-r",
        "--role",
        type=str,
        required=False,
        choices=["superadmin", "admin", "user", "guest"],
        help="The user role",
    )
    parser.add_argument(
        "-c",
        "--cell_limit",
        type=int,
        required=False,
        help="The maximum number of cells a user can process",
    )
    parser.add_argument(
        "-n",
        "--process_num_limit",
        type=int,
        required=False,
        help="The maximum number of processes the user can execute"
        " in a single process chain",
    )
    parser.add_argument(
        "-t",
        "--process_time_limit",
        type=int,
        required=False,
        help="The maximum number seconds a single process is allowed to run",
    )
    parser.add_argument(
        "-d",
        "--datasets",
        type=str,
        required=False,
        help="The datasets the user is allowed to access."
        " Format: location/mapset,location/mapset",
    )
    parser.add_argument(
        "-m",
        "--modules",
        type=str,
        required=False,
        help="A list of modules the user is allowed to access."
        " Format: module,module,module",
    )

    args = parser.parse_args()

    conf = Configuration()
    try:
        conf.read()
    except Exception:
        pass

    server = conf.REDIS_SERVER_URL
    port = conf.REDIS_SERVER_PORT
    if conf.REDIS_SERVER_PW:
        redis_password = conf.REDIS_SERVER_PW
    else:
        redis_password = None

    if args.server:
        server = args.server
    if args.port:
        port = args.port
    if args.redis_password:
        redis_password = args.redis_password
    redis_user_interface.connect(
        host=server,
        port=port,
        password=redis_password,
    )

    # CREATE ############################
    if args.action == "create":
        if args.user_id is None:
            sys.stderr.write("You need to provide a user id\n")
            return

        if args.user_group is None:
            sys.stderr.write("You need to provide a user group\n")
            return

        if args.password is None:
            sys.stderr.write("You need to provide a password\n")
            return

        if args.role is None:
            sys.stderr.write("You need to provide a user role\n")
            return

        user = ActiniaUser(user_id=args.user_id, user_group=args.user_group)

        if user.exists() == 1:
            sys.stderr.write(
                "Unable to create the user <%s> the user exists\n"
                % args.user_id
            )
            return

        # Set the credentials
        set_user_credentials(user, args, args.action)

        if user.commit() is True:
            sys.stderr.write(
                "Created user <%s> in group <%s>\n"
                % (args.user_id, args.user_group)
            )
            sys.stdout.write(str(user))
            return
        else:
            sys.stderr.write("Unable to create the user <%s>\n" % args.user_id)
            return
    # UPDATE ############################
    elif (
        args.action == "update"
        or args.action == "update_add"
        or args.action == "update_rm"
        or args.action == "update_rm_location"
    ):
        if args.user_id is None:
            sys.stderr.write("You need to provide a user id\n")
            return

        user = ActiniaUser(user_id=args.user_id)

        if user.exists() == 0:
            sys.stderr.write(
                "Unable to update the user <%s> the user doesn't exist\n"
                % args.user_id
            )
            return

        user.read_from_db()

        # Set the credentials
        set_user_credentials(user, args, args.action)

        if user.update() is True:
            sys.stderr.write("Updated user <%s>\n" % args.user_id)
            sys.stdout.write(str(user))
            return
        else:
            sys.stderr.write("Unable to update the user <%s>\n" % args.user_id)
            return

    # DELETE ############################
    elif args.action == "delete":
        user = ActiniaUser(args.user_id)

        if user.exists() == 1:
            if user.delete() is True:
                sys.stderr.write("User <%s> deleted\n" % args.user_id)
                return
            else:
                sys.stderr.write("Unable to delete user <%s>\n" % args.user_id)
                return
        else:
            sys.stderr.write("User <%s> does not exist\n" % args.user_id)

    # SHOW ##############################
    elif args.action == "show":
        user = ActiniaUser(args.user_id)

        if user.exists() == 1:
            sys.stdout.write(str(user))
        else:
            sys.stderr.write("User <%s> does not exist\n" % args.user_id)

    # LIST ##############################
    elif args.action == "list":
        user = ActiniaUser(args.user_id)
        pprint.pprint(user.list_all_users())

    # PWCHECK ###########################
    elif args.action == "pwcheck":
        if args.user_id is None:
            sys.stderr.write("You need to provide a user id\n")
            return

        if args.password is None:
            sys.stderr.write("You need to provide a password\n")
            return

        if verify_password(args.user_id, args.password) is False:
            sys.stderr.write("Wrong username or password\n")
        else:
            sys.stderr.write("Password and user name are correct\n")


if __name__ == "__main__":
    main()



#!flask/bin/python
# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Soeren Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Dummy server to response HTTP code 200 to a POST call from an actinia webhook
"""
import argparse
import psutil
from flask import Flask, make_response, jsonify, request, json
from multiprocessing import Process
from werkzeug.exceptions import BadRequest, UnsupportedMediaType

from pprint import pprint

__license__ = "GPLv3"
__author__ = "Soeren Gebbert, Anika Weinmann, Carmen Tawalika"
__copyright__ = "Copyright 2016-2023, mundialis GmbH & Co. KG"
__maintainer__ = "mundialis GmbH & Co. KG"


flask_app = Flask(__name__)


@flask_app.route("/webhook/finished", methods=["GET", "POST"])
def finished():
    port = request.server[1]
    try:
        pprint(json.loads(request.get_json()))
    except BadRequest:
        pass
    except UnsupportedMediaType:
        pass
    except TypeError:
        pass
    sp = Process(target=shutdown_server, args=(port,))
    sp.start()
    return make_response(jsonify("Server shutting down..."), 200)


@flask_app.route("/webhook/update", methods=["GET", "POST"])
def update():
    try:
        pprint(json.loads(request.get_json()))
    except BadRequest:
        pass
    except UnsupportedMediaType:
        pass
    except TypeError:
        pass
    return make_response(jsonify("OK"), 200)


def shutdown_server(port):
    for proc in psutil.process_iter():
        if (
            proc.name() == "webhook-server-"
            and proc.as_dict()["connections"]
            and proc.as_dict()["connections"][0].laddr.port == port
        ):
            proc.kill()


@flask_app.route("/shutdown", methods=["GET"])
def shutdown():
    port = request.server[1]
    sp = Process(target=shutdown_server, args=(port,))
    sp.start()
    return make_response(jsonify("Server shutting down..."), 200)


def main():
    parser = argparse.ArgumentParser(
        description="Start a webhook server that exposes GET/POST endpoints "
        "which returns HTTP code 200 if called. The endpoints are: "
        " - /webhook/finished for finished callbacks "
        " - /webhook/update for status update callbacks"
    )

    parser.add_argument(
        "--host",
        type=str,
        required=False,
        default="0.0.0.0",
        help="The IP address that should be used for the webhook server",
    )

    parser.add_argument(
        "--port",
        type=int,
        required=False,
        default=5005,
        help="The port that should be used for the webhook server",
    )

    args = parser.parse_args()

    flask_app.run(host=args.host, port=args.port)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Tests: Async process test case
"""
import unittest

# import pytest
from flask.json import loads as json_loads, dumps as json_dumps
import requests
from jinja2 import Template
import os
import time

try:
    from .test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX
except ModuleNotFoundError:
    from test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX

__license__ = "GPLv3"
__author__ = "Anika Weinmann"
__copyright__ = "Copyright 2022, mundialis GmbH & Co. KG"
__maintainer__ = "mundialis GmbH & Co. KG"

port = "5006"

pc = {
    "version": 1,
    "webhooks": {
        "finished": f"http://0.0.0.0:{port}/webhook/finished",
        "update": f"http://0.0.0.0:{port}/webhook/update",
    },
    "list": [
        {"id": "1", "exe": "/bin/sleep", "params": ["{{ sleep }}"]},
        {
            "id": "2",
            "module": "g.region",
            "inputs": [
                {"param": "raster", "value": "elevation@PERMANENT"},
                {"param": "res", "value": "10000"},
            ],
            "flags": "p",
            "verbose": True,
        },
    ],
}


def startWebhook(sleeptime=10):
    """
    Starts a webhook server which returns status code 200 for /webhook/update
    and /webhook/finished endpoints and can be shutdown by endpoint /shutdown.
    """
    inputlist = [
        "webhook-server",
        "--host",
        "0.0.0.0",
        "--port",
        port,
        "&",
    ]
    os.system(" ".join(inputlist))
    time.sleep(sleeptime)
    resp = requests.get(f"http://0.0.0.0:{port}/webhook/finished")
    return resp.status_code


def startBrokenWebhook(sleeptime=10):
    """
    Starts a webhook server which returns status code 200 for /webhook/update
    endpoint and by endpoint /webhook/finished it shuts down itself to test the
    etries if the webhook server can not be reached or return a server error
    code (500 - 599).
    """
    inputlist = [
        "webhook-server-broken",
        "--host",
        "0.0.0.0",
        "--port",
        port,
        "&",
    ]
    os.system(" ".join(inputlist))
    time.sleep(sleeptime)
    resp = requests.get(f"http://0.0.0.0:{port}/webhook/update")
    return resp.status_code


class WebhookTestCase(ActiniaResourceTestCaseBase):
    def setUp(self):
        # start a webhook server before the normal setUp
        status_code = startWebhook(10)
        self.assertEqual(status_code, 200, "Webhook server is not running")
        super().setUp()

    def tearDown(self):
        # shutting down the webhook server
        resp = requests.get(f"http://0.0.0.0:{port}/shutdown")
        self.assertEqual(
            resp.status_code, 200, "Webhook server is not shutting down"
        )

    def poll_job(self, resp_data):
        # poll an actinia job
        rv_user_id = resp_data["user_id"]
        rv_resource_id = resp_data["resource_id"]
        time.sleep(10)
        rv2 = self.server.get(
            URL_PREFIX + "/resources/%s/%s" % (rv_user_id, rv_resource_id),
            headers=self.admin_auth_header,
        )
        resp_data2 = json_loads(rv2.data)
        return resp_data2

    # @pytest.mark.dev
    # def test_started_webhook(self):
    #     """Test the started webhook via a actinia process."""
    #     # time.sleep(10)
    #     # import pdb; pdb.set_trace()
    #     resp = None
    #     while resp is None or resp.status_code != 200:
    #         resp = requests.get(f"http://0.0.0.0:{port}/webhook/finished")
    #         time.sleep(3)
    #     tm = Template(json_dumps(pc))
    #     rv = self.server.post(
    #         URL_PREFIX + "/locations/nc_spm_08/processing_async_export",
    #         headers=self.admin_auth_header,
    #         data=tm.render(sleep=1),
    #         content_type="application/json",
    #     )
    #     self.waitAsyncStatusAssertHTTP(rv, headers=self.admin_auth_header)

    def test_finished_webhook_retries(self):
        """Test the retry if the webhook can not be reached or returns a server
        error.
        """
        tm = Template(json_dumps(pc))
        rv = self.server.post(
            URL_PREFIX + "/locations/nc_spm_08/processing_async_export",
            headers=self.admin_auth_header,
            data=tm.render(sleep=30),
            content_type="application/json",
        )

        # poll job and check if it is running
        resp_data = json_loads(rv.data)
        status = resp_data["status"]
        while status == "accepted":
            resp_data2 = self.poll_job(resp_data)
            status = resp_data2["status"]
        self.assertEqual(status, "running", "Actinia job is not 'running'!")

        # shutdown webhook and start broken finished webhook
        webhook_resp = requests.get(f"http://0.0.0.0:{port}/shutdown")
        self.assertEqual(
            webhook_resp.status_code, 200, "Shutdown of webhook failed!"
        )
        time.sleep(2)
        status_code = startBrokenWebhook(10)
        self.assertEqual(
            status_code, 200, "Broken webhook server is not running"
        )

        broken_server_is_running = True
        while broken_server_is_running is True:
            try:
                resp = requests.get(f"http://0.0.0.0:{port}/webhook/update")
                if resp.status_code == 500:
                    broken_server_is_running = False
            except Exception:
                broken_server_is_running = False

        # restart webhook server
        status_code = startWebhook(10)
        self.assertEqual(status_code, 200, "Webhook server restart failed")

        # check if job is finished
        resp_data2 = self.poll_job(resp_data)
        status = resp_data2["status"]
        self.assertEqual(status, "finished", "Actinia job is not finished")


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2021 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Process chain
"""
from flask_restful_swagger_2 import Schema
from copy import deepcopy
from actinia_api import URL_PREFIX

__license__ = "GPLv3"
__author__ = "Sören Gebbert, Carmen Tawalika"
__copyright__ = (
    "Copyright 2016-2021, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


SUPPORTED_EXPORT_FORMATS = [
    "COG",
    "GTiff",
    "GPKG",
    "SQLite",
    "GML",
    "GeoJSON",
    "ESRI_Shapefile",
    "CSV",
    "TXT",
    "PostgreSQL",
]


class IOParameterBase(Schema):
    """This is the input parameter of a GRASS GIS module definition"""

    type = "object"
    properties = {
        "param": {
            "type": "string",
            "description": "The name of a GRASS GIS module parameter like "
            "*map* or *elevation*. ",
        },
        "value": {
            "type": "string",
            "description": "The value of the GRASS GIS module parameter. "
            "Raster, vector and STDS inputs must contain the "
            "mapset name in their id: *slope@PERMANENT*, if "
            "they are not located in the working mapset. "
            "Do not contain the mapset name in map names that are "
            "processed, since the mapsets are generated on demand "
            "using random names. "
            "Outputs are not allowed to contain mapset names."
            "Files that are created in the process chain to exchange "
            "data can be specified using the *$file::unique_id* "
            "identifier. The **unique_id** will be replaced with a "
            "temporary file name, that is available in the whole "
            "process chain at runtime. The **unique_id**  is the "
            "identifier that can be used by different modules in "
            "a process chain to access the same temporary file or "
            "to prepare it for export.",
        },
    }
    description = (
        "Parameter definition of a GRASS GIS module that should be executed "
        "in the actinia environment. Parameters can be of type input or "
        "output. A GRASS GIS module will be usually called like: "
        "<p>g.region raster=elevation30m@PERMANENT</p> "
        "The GRASS GIS module *g.region* parameter *raster* has the value "
        "*elevation30m@PERMANENT*. "
        "This is reflected by the *param* and *value* properties that can "
        "specify input and output "
        "parameters."
    )
    required = ["param", "value"]
    example = {"param": "file", "value": "$file::ascii_points"}


class InputParameter(IOParameterBase):
    """
    This is the import definition of a input parameter of a GRASS GIS module
    definition
    """

    properties = deepcopy(IOParameterBase.properties)
    properties["import_descr"] = {
        "type": "object",
        "description": "Definition of sources to be imported as raster or "
        "vector map layer.",
        "properties": {
            "type": {
                "type": "string",
                "description": "The type of the input that should be "
                "downloaded and imported. In case of raster or vector types "
                "a download URL must be provided as source using "
                "http, https or ftp protocols. In case of sentinel2 "
                " scenes the scene name and the band must be provided. "
                "The Landsat approach is different. <br><br>"
                "In case a Landsat scene is requested, all "
                "bands will be download, in the target location imported"
                " and an atmospheric correction is applied. The "
                "atmospheric correction must be specified. The resulting"
                " raster map layers have a specific name scheme, that "
                "is independent from the provided map name in the "
                "process description. The name scheme is always: "
                "<p>\<landsat scene id\>_\<atcor\>.\<band\></p>"
                "For example, if the scene <p>LT52170762005240COA00</p> "
                "was requested, the resulting name for the DOS1 "
                "atmospheric corrected band 1 would be: "
                "<p>LT52170762005240COA00_dos1.1</p>."
                "For the DOS1 atmospheric corrected band 2 it "
                "would be: <p>LT52170762005240COA00_dos1.2</p> "
                "and so on. "
                "All other process steps must use these raster map layer"
                " names to refer to the imported Landsat bands. "
                "Use the file option to download any kind of files that "
                "should be processed by a grass gis module. ",
                "enum": [
                    "raster",
                    "vector",
                    "landsat",
                    "sentinel2",
                    "postgis",
                    "file",
                    "stac",
                ],
            },
            "sentinel_band": {
                "type": "string",
                "description": "The band of the sentinel2 scene that should be"
                " imported",
                "enum": [
                    "B01",
                    "B02",
                    "B03",
                    "B04",
                    "B05",
                    "B06",
                    "B07",
                    "B08",
                    "B8A",
                    "B09",
                    "B10",
                    "B11",
                    "B12",
                ],
            },
            "landsat_atcor": {
                "type": "string",
                "description": "The atmospheric correction that should be "
                "applied to the landsat scene",
                "enum": [
                    "uncorrected",
                    "dos1",
                    "dos2",
                    "dos2b",
                    "dos3",
                    "dos4",
                ],
            },
            "vector_layer": {
                "type": "string",
                "description": "The name of the layer that should be imported "
                "from the vector file or postGIS database",
            },
            "source": {
                "type": "string",
                "description": "The input source that may be a landsat scene "
                "name, a sentinel2 scene name, a postGIS database string,"
                "a stac collection ID or an URL that points "
                "to an accessible raster or vector file. "
                "A HTTP, HTTPS or FTP connection must be "
                "specified in case of raster or vector types. "
                "In this case the source string must contain the "
                "protocol that will used for connection: http:// or "
                "https:// or ftp://. PostGIS vector layer can be "
                "imported by defining a database string as source and "
                "a layer name.",
            },
            "semantic_label": {
                "type": "string",
                "description": "Refers to the common names used to"
                "call the bands of an image,"
                "for example: red, blue, nir, swir"
                "However, this property also accepts the band name"
                "such as B1, B8 etc. The semantic labeling should match"
                "the labels register in the stac collection.",
                "example": "red, blue, nir or B1, B2, B8A",
            },
            "extent": {
                "type": "string",
                "description": "Spatio-temporal constraint defined by the user"
                "throughout bbox and interval concept.",
                "enum": ["spatial", "temporal"],
                "example": {
                    "spatial": {"bbox": [[30.192, -16.369, 42.834, -0.264]]},
                    "temporal": {"interval": [["2021-09-09", "2021-09-12"]]},
                },
            },
            "filter": {
                "type": "string",
                "description": "Constrain in any other property"
                "or metadata.",
                "example": {"query": {"eo:cloud_cover": {"lt": 50}}},
            },
            "resample": {
                "type": "string",
                "description": "Resampling method to use for reprojection of "
                "raster map (default: nearest).",
                "enum": [
                    "nearest",
                    "bilinear",
                    "bicubic",
                    "lanczos",
                    "bilinear_f",
                    "bicubic_f",
                    "lanczos_f",
                ],
            },
            "resolution": {
                "type": "string",
                "description": "Resolution of output raster map. Estimated, "
                "user-specified or current region resolution "
                "(default: estimated).",
                "enum": ["estimated", "value", "region"],
            },
            "resolution_value": {
                "type": "string",
                "description": "Resolution of output raster map (use with "
                'option "resolution": "value") in units of the target '
                "coordinate reference system, not in map units. Must "
                "be convertible to float.",
                "example": {
                    "import_descr": {
                        "source": "https://example.tiff",
                        "type": "raster",
                        "resample": "bilinear_f",
                        "resolution": "value",
                        "resolution_value": "100",
                    }
                },
            },
            "basic_auth": {
                "type": "string",
                "description": "User name and password for basic HTTP, HTTPS "
                "and FTP authentication of the source connection. The user "
                "name and password must be separated by a colon: "
                "username:password",
            },
        },
    }

    required = deepcopy(IOParameterBase.required)
    description = deepcopy(IOParameterBase.description)
    example = {
        "import_descr": {
            "source": "LT52170762005240COA00",
            "type": "landsat",
            "landsat_atcor": "dos1",
        },
        "param": "map",
        "value": "ignored",
    }


class OutputParameter(IOParameterBase):
    """This is the output parameter of a GRASS GIS module definition"""

    type = "object"
    properties = deepcopy(IOParameterBase.properties)
    properties["export"] = {
        "type": "object",
        "properties": {
            "format": {
                "type": "string",
                "description": "The format of the output file in case of "
                "raster, strds, vector layer or text file export. Raster layer"
                " export supports only GTiff and COG format, STRDS layer "
                "export supports only GTiff format and all other formats are "
                "vector layer export formats. If the *PostgeSQL* format "
                "was chosen, a postgis database string *dbstring* "
                "must be provided  so that the GRASS GIS module "
                "*v.out.ogr knows to which PostgreSQL database it "
                "should be connect. The name of the output layer can be "
                "specified as *output_layer* for PostgreSQL database "
                "exports. Some GRASS GIS modules allow the export of "
                "text files. These files can be exported and provided "
                "as downloadable link as well.",
                "enum": SUPPORTED_EXPORT_FORMATS,
            },
            "type": {
                "type": "string",
                "description": "The type of the output. In case of the *file* "
                "option only text files are supported for export. "
                "In addition the *$file::unique_id* option must "
                "be used in the value parameter to export a file "
                "that was generated by a GRASS GIS module. Exported "
                "text and vector files will always be compressed "
                "with zip.",
                "enum": ["raster", "vector", "file", "strds"],
            },
            "dbstring": {
                "type": "string",
                "description": "The database string to be used to connect to a"
                " PostgreSQL database for vector export.",
            },
            "output_layer": {
                "type": "string",
                "description": "Name for output PostgreSQL layer. If not "
                "specified, GRASS GIS vector map layer name is used.",
            },
        },
        "description": "The raster, vector or text file export parameter.",
        "required": ["format", "type"],
        "example": {
            "format": "PostgreSQL",
            "type": "vector",
            "dbstring": "PG:host=localhost dbname=postgis user=postgres",
            "output_layer": "roads",
        },
    }
    properties["metadata"] = {
        "type": "object",
        "properties": {
            "format": {
                "type": "string",
                "description": "Format of the metadata file. Only STAC is "
                "supported. "
                "The STAC item builder works just on raster export file."
                "These files are accessible through a STAC Catalog"
                "and each export is stored as STAC Item",
                "enum": ["STAC"],
            },
        },
        "description": "The STAC file export parameter.",
        "required": ["format", "type"],
        "example": {"format": "STAC"},
    }
    required = deepcopy(IOParameterBase.required)
    description = deepcopy(IOParameterBase.description)
    example = {
        "param": "slope",
        "value": "elev_10m_slope",
        "export": {"type": "raster", "format": "GTiff"},
    }


class StdoutParser(Schema):
    """This is the schema of the stdout output parser for GRASS GIS modules"""

    type = "object"
    properties = {
        "id": {
            "type": "string",
            "description": "The unique id that is used to identify the "
            "parsed output in the result dictionary.",
        },
        "format": {
            "type": "string",
            "description": "The stdout format to be parsed.",
            "enum": ["table", "list", "kv", "json"],
        },
        "delimiter": {
            "type": "string",
            "description": "The delimiter that should be used to parse table,"
            " list and key/value module output. Many GRASS "
            'GIS  modules use by default "|" in tables and '
            '"=" in key/value pairs. A new line "\\n" is '
            "always the delimiter between rows in the output.",
        },
    }
    required = ["id", "format", "delimiter"]
    description = (
        "Use this parameter to automatically parse the output of GRASS GIS "
        "modules and convert the output into tables, lists or key/value pairs "
        "in the result section of the response."
        "If the property type is set to *table*, *list* or *kv* then "
        "the stdout of the current command will be parsed and "
        "the result of the parse operation will be added to the "
        "result dictionary using the provided id as key. GRASS GIS modules "
        "produce regular output. Many modules have the flag *-g* to "
        "create key value pairs as stdout output. Other create a list of "
        "values or a table with/without header."
    )
    example = {"id": "stats", "format": "table", "delimiter": "|"}


class GrassModule(Schema):
    """
    The definition of a GRASS GIS module and its inputs, outputs and flags
    """

    type = "object"
    properties = {
        "id": {
            "type": "string",
            "description": "A unique id to identify the module call in the "
            "process chain to reference its stdout and stderr "
            "output as stdin in other modules.",
        },
        "module": {
            "type": "string",
            "description": "The name of the GRASS GIS module (r.univar, "
            "r.slope.aspect, v.select, ...) that should be "
            'executed. Use as module names "importer" or '
            '"exporter" to import or export raster layer, '
            "vector layer or other file based data without "
            "calling a GRASS GIS module.",
        },
        "inputs": {
            "type": "array",
            "items": InputParameter,
            "description": "A list of input parameters of a GRASS GIS module.",
        },
        "outputs": {
            "type": "array",
            "items": OutputParameter,
            "description": "A list of output parameters of a GRASS GIS "
            "module.",
        },
        "flags": {
            "type": "string",
            "description": "The flags that should be set for the GRASS GIS "
            "module.",
        },
        "stdin": {
            "type": "string",
            "description": "Use the stdout output of a GRASS GIS module or "
            "executable of the process chain as input for this "
            "module. Refer to the module/executable output "
            'as *id::stderr* or *id::stdout*, the "id" is '
            "the unique identifier of a GRASS GIS module "
            "definition.",
        },
        "stdout": StdoutParser,
        "overwrite": {
            "type": "boolean",
            "description": "Set True to overwrite existing data.",
        },
        "verbose": {
            "type": "boolean",
            "description": "Set True to activate verbosity output of the "
            "module.",
        },
        "superquiet": {
            "type": "boolean",
            "description": "Set True to silence the output of the " "module.",
        },
        "interface-description": {
            "type": "boolean",
            "description": "Set True to print interface "
            "description and exit.",
        },
    }
    required = ["id", "module"]
    description = (
        "The definition of a single GRASS GIS module and its inputs, outputs "
        "and flags. This module will be run in a location/mapset environment "
        "and is part of a process chain. The stdout and stderr output of "
        "modules that were run before this module in the process chain can be "
        "used as stdin for this module. The stdout of a module can be "
        "automatically transformed in list, table or key/value JSON "
        "representations in the HTTP response."
    )
    example = {
        "module": "r.slope.aspect",
        "id": "r_slope_aspect_1",
        "inputs": [
            {
                "import_descr": {
                    "source": "https://storage.googleapis.com/graas-geodata/"
                    "elev_ned_30m."
                    "tif",
                    "type": "raster",
                },
                "param": "raster",
                "value": "elev_ned_30m_new",
            }
        ],
        "outputs": [
            {
                "export": {"format": "GTiff", "type": "raster"},
                "param": "slope",
                "value": "elev_ned_30m_new_slope",
            }
        ],
        "flags": "a",
    }


class Executable(Schema):
    """The definition of a Linux executable and its parameters"""

    type = "object"
    properties = {
        "id": {
            "type": "string",
            "description": "A unique id to identify the executable in the "
            "process chain to reference its stdout and stderr output as "
            "stdin in other modules.",
        },
        "exe": {
            "type": "string",
            "description": "The name of the Linux executable that should be "
            "executed as part of the process chain.",
        },
        "params": {
            "type": "array",
            "items": {"type": "string"},
            "description": "A list of input parameters of a GRASS GIS module.",
        },
        "stdin": {
            "type": "string",
            "description": "Use the output of a GRASS GIS module or executable"
            " in of the process chain as input for this module. "
            "Refer to the module/executable output as id::stderr "
            'or id::stdout, the "id" is the unique identifier '
            "of a GRASS GIS module.",
        },
    }
    required = ["id", "exe"]
    description = (
        "The definition of a Linux executable and its parameters. "
        "The stdout and stderr output of modules/executable "
        "that were run before this module in the process "
        "chain can be used as stdin for this module."
    )
    example = {"id": "cat_1", "exe": "/bin/cat", "params": []}


class Webhooks(Schema):
    """The definition of finished and update webhooks"""

    type = "object"
    properties = {
        "update": {
            "type": "string",
            "description": "Specify a HTTP(S) GET/POST endpoint that should be"
            " called when a status update is available while the "
            "process chain is executed. The actinia JSON status "
            "response will be send as JSON content to the POST "
            "endpoint for each status update until the process "
            "finished. The GET endpoint, that must be available "
            "by the same URL as the POST endpoint, will be used "
            "to check if the webhook endpoint is available.",
        },
        "finished": {
            "type": "string",
            "description": "Specify a HTTP(S) GET/POST endpoint that should "
            "be called when the process chain was executed "
            "successful or unsuccessfully. The actinia JSON "
            "response will be send as JSON content to the POST "
            "endpoint after processing finished. The GET "
            "endpoint, that must be available by the same URL "
            "as the POST endpoint, will be used to check if "
            "the webhook endpoint is available.",
        },
    }
    required = ["finished"]
    description = (
        "Specify HTTP(S) GET/POST endpoints that should be called when the "
        "process chain was executed successful or unsuccessfully (finished) or"
        " when a status/progress update is available (update). The actinia "
        "JSON response will be send as JSON content to the POST endpoints "
        "after processing finished or the status was updated. The GET "
        "endpoints, that must be available by the same URL as the POST "
        "endpoints (update/finished), will be used to check if the webhooks "
        "endpoints are available. "
        "The finished endpoint is mandatory, the update endpoint is optional."
    )
    example = {
        "update": f"http://business-logic.company.com{URL_PREFIX}/"
        "actinia-update-webhook",
        "finished": f"http://business-logic.company.com{URL_PREFIX}/"
        "actinia-finished-webhook",
    }


class ProcessChainModel(Schema):
    """Definition of the actinia process chain that includes GRASS GIS modules
    and common Linux commands
    """

    type = "object"
    properties = {
        "version": {
            "type": "string",
            "default": "1",
            "description": "The version string of the process chain",
        },
        "list": {
            "type": "array",
            "items": GrassModule,
            "description": "A list of process definitions that should be "
            "executed in the order provided by the list.",
        },
        "webhooks": Webhooks,
    }
    required = ["version", "list"]
    example = {
        "list": [
            {
                "module": "g.region",
                "id": "g_region_1",
                "inputs": [
                    {
                        "import_descr": {
                            "source": "https://storage.googleapis.com/"
                            "graas-geodata/elev_ned_30m.tif",
                            "type": "raster",
                        },
                        "param": "raster",
                        "value": "elev_ned_30m_new",
                    }
                ],
                "flags": "p",
            },
            {
                "module": "r.slope.aspect",
                "id": "r_slope_aspect_1",
                "inputs": [
                    {"param": "elevation", "value": "elev_ned_30m_new"}
                ],
                "outputs": [
                    {
                        "export": {"format": "GTiff", "type": "raster"},
                        "param": "slope",
                        "value": "elev_ned_30m_new_slope",
                    }
                ],
                "flags": "a",
            },
            {
                "module": "r.univar",
                "id": "r_univar_1",
                "inputs": [
                    {
                        "import_descr": {
                            "source": "LT52170762005240COA00",
                            "type": "landsat",
                            "landsat_atcor": "dos1",
                        },
                        "param": "map",
                        "value": "LT52170762005240COA00_dos1.1",
                    }
                ],
                "stdout": {"id": "stats", "format": "kv", "delimiter": "="},
                "flags": "a",
            },
            {
                "module": "exporter",
                "id": "exporter_1",
                "outputs": [
                    {
                        "export": {"format": "GTiff", "type": "raster"},
                        "metadata": {
                            "format": "STAC",
                            "type": "metadata",
                            "output_layer": "stac_result",
                        },
                        "param": "map",
                        "value": "LT52170762005240COA00_dos1.1",
                    }
                ],
            },
            {
                "id": "ascii_out",
                "module": "r.out.ascii",
                "inputs": [
                    {"param": "input", "value": "elevation@PERMANENT"},
                    {"param": "precision", "value": "0"},
                ],
                "stdout": {
                    "id": "elev_1",
                    "format": "table",
                    "delimiter": " ",
                },
                "flags": "h",
            },
            {
                "id": "ascii_export",
                "module": "r.out.ascii",
                "inputs": [{"param": "input", "value": "elevation@PERMANENT"}],
                "outputs": [
                    {
                        "export": {"type": "file", "format": "TXT"},
                        "param": "output",
                        "value": "$file::out1",
                    }
                ],
            },
            {
                "id": "raster_list",
                "module": "g.list",
                "inputs": [{"param": "type", "value": "raster"}],
                "stdout": {
                    "id": "raster",
                    "format": "list",
                    "delimiter": "\n",
                },
            },
            {
                "module": "r.what",
                "id": "r_what_1",
                "verbose": True,
                "flags": "nfic",
                "inputs": [
                    {"param": "map", "value": "landuse96_28m@PERMANENT"},
                    {
                        "param": "coordinates",
                        "value": "633614.08,224125.12,632972.36,225382.87",
                    },
                    {"param": "null_value", "value": "null"},
                    {"param": "separator", "value": "pipe"},
                ],
                "stdout": {
                    "id": "sample",
                    "format": "table",
                    "delimiter": "|",
                },
            },
        ],
        "webhooks": {
            "update": f"http://business-logic.company.com{URL_PREFIX}/"
            "actinia-update-webhook",
            "finished": f"http://business-logic.company.com{URL_PREFIX}/"
            "actinia-finished-webhook",
        },
        "version": "1",
    }



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Redis server resource logging interface
"""

from actinia_core.core.common.redis_base import RedisBaseInterface

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Sören Gebbert"
__email__ = "soerengebbert@googlemail.com"


class RedisResourceInterface(RedisBaseInterface):
    """
    The Redis resource database interface
    """

    # The database to store the long pending resource status and results
    resource_id_prefix = "RESOURCE-ID::"
    resource_id_termination_prefix = "RESOURCE-ID-TERMINATION::"

    def __init__(self):
        """
        The resource database stores information about a resource that was
        created by a GRASS process or that is in state of creation. The GRASS
        process will continually update the state of the resource by altering
        the resource information.

        Other processes can access the state of the resource and deliver it to
        clients, independently which server process creates and updates the
        resource.

        Database entries are based on setex and have an expiration time of
        8640000 seconds (100 days) by default. This can also be set using the
        Actinia Core config file.

        The following scheme should be used::

            Client         GRASS Server              Resource DB
               |                |                         |
               |---------> Submit job     ---------> Create entry set accepted
               |<--------- Resource id                    |
               |                |                         |
               |           Start job      ---------> Update entry set running
               |                |                         |
               |---------> Status?                        |
               |           Request status ---------> Read entry
               |<--------- Deliver status <--------- Return entry
               |                |                         |
               |           Run job        ---------> Update entry set running
               |                |                         |
               |           Finish job     ---------> Update entry set finished
               |                |                         |
               |---------> Status?                        |
               |           Request status ---------> Read entry
               |<--------- Deliver status <--------- Return entry
               |                |                         |
               V                V                         V

            # Termination of a job

            Client         GRASS Server              Resource DB
               |                |                         |
               |---------> Submit job     ---------> Create entry
               |<--------- Resource id                    |
               |                |                         |
               |           Start job      ---------> Update entry set running
               |                |                         |
               |           Request status ---------> Read termination entry
               |           Run job        ---------> Update entry set running
               |                |                         |
               |-----------------------------------> Set termination entry
               |                |                         |
               |           Request status ---------> Read termination entry
               |           Terminate job  ---------> Update entry set
                                                     terminated
               |                |                         |
               |---------> Status?                        |
               |           Request status ---------> Read entry
               |<--------- Deliver status <--------- Return entry
               |                |                         |
               V                V                         V

        A GRASS server that delivers the job status may not be the same
        server that computes the results.

        The client does not set the termination entry directly,
        a GRASS REST process will do this.

        The GRASS process runner that updates the resource entries will
        check periodically for termination entries. If a termination
        entry exists, the job will be terminated and all allocated resources
        (RAM, HD) will be freed.

        """
        RedisBaseInterface.__init__(self)

    def set(self, resource_id, resource_entry, expiration=864000):
        """Set or update a resource entry

        Args:
            resource_id (str): The unique id of the resource
            resource_entry (str): The entry that should be put in the database
            expiration (int): The time in seconds when this resource should
                              expire

        """
        return self.redis_server.setex(
            self.resource_id_prefix + resource_id, expiration, resource_entry
        )

    def set_termination(self, resource_id, expiration=3600):
        """Set or update a resource termination entry

        The running job will check for termination periodically and will
        terminate the job if an entry exists.

        Args:
            resource_id (str): The unique id of the resource that should be
                               terminated
            expiration (int): The time in seconds when this resource should
                              expire

        """
        return self.redis_server.setex(
            self.resource_id_termination_prefix + resource_id, expiration, 1
        )

    def get(self, resource_id):
        """Get the resource entry if exists

        Args:
            resource_id (str): The unique id of the resource

        Returns:
            str:
            The resource entry or None
        """
        value = self.redis_server.get(self.resource_id_prefix + resource_id)
        return value

    def get_keys_from_pattern(self, resource_id_pattern):
        """Get all keys of a resource_id_pattern

        Args:
            resource_id_pattern (str): The pattern to the resources

        Returns:
            list:
            The list of the matching kesy
        """
        values = self.redis_server.scan_iter(
            self.resource_id_prefix + resource_id_pattern
        )
        resource_keys = list()
        for val in values:
            resource_keys.append(
                val.decode("utf-8").replace(self.resource_id_prefix, "")
            )
        return resource_keys

    def get_list(self, regexpr):
        """Get a list of resource entries if exists

        Args:
            regexpr (str): A regular expression

        Returns:
            list:
            A list of resource entries
        """
        key_list = self.redis_server.keys(self.resource_id_prefix + regexpr)
        resource_list = []
        if key_list:
            for key in key_list:
                value = self.redis_server.get(key.decode())
                resource_list.append(value)
        return resource_list

    def get_termination(self, resource_id):
        """Get the resource termination entry if exists

        The running job will check for termination periodically and will
        terminate the job if an entry exists.

        Args:
            resource_id (str): The unique id of the resource

        Returns:
            bool:
            True or False
        """
        return bool(
            self.redis_server.get(
                self.resource_id_termination_prefix + resource_id
            )
        )

    def get_termination_list(self, regexpr):
        """Get a list of termination resource entries if exists

        Args:
            regexpr (str): A regular expression

        Returns:
            list:
            A list of resource entries
        """
        term_key_list = self.redis_server.keys(
            self.resource_id_termination_prefix + regexpr
        )

        resource_list = []
        if term_key_list:
            for key in term_key_list:
                resource_list.append(key.decode())

        return resource_list

    def delete(self, resource_id):
        """Delete a resource entry

        Args:
            resource_id (str): The unique id of the resource

        """
        return self.redis_server.delete(self.resource_id_prefix + resource_id)

    def delete_termination(self, resource_id):
        """Delete a termination resource entry

        Args:
            resource_id (str): The unique id of the resource

        """
        return self.redis_server.delete(
            self.resource_id_termination_prefix + resource_id
        )


# Create the Redis interface instance
# redis_resource_interface = RedisResourceInterface()


def test_resource_entries(r):
    resource_id = "abcdefg"

    # Remove the resource entry if it exist
    r.delete(resource_id)

    ret = r.set(resource_id, "JSON file", expiration=20)
    if ret != 1:
        raise Exception("set does not work")

    ret = r.set(resource_id, "JSON file update", expiration=20)
    if ret != 1:
        raise Exception("set does not work")

    ret = r.get(resource_id)
    if ret != "JSON file update":
        raise Exception("get does not work")

    ret = r.set_termination(resource_id, expiration=20)
    if ret != 1:
        raise Exception("set_termination does not work")

    ret = r.get_termination(resource_id)
    if ret is not True:
        raise Exception("get_termination does not work")

    ret = r.delete(resource_id)
    if ret != 1:
        raise Exception("delete does not work")

    ret = r.get(resource_id)
    if ret is not None:
        raise Exception("get does not work")

    ret = r.delete_termination(resource_id)
    if ret != 1:
        raise Exception("delete_termination does not work")

    ret = r.get_termination(resource_id)
    if ret is not False:
        raise Exception("get_termination does not work")

    # Try the keys implementation

    r.delete(resource_id)

    ret = r.set(resource_id + "_1", "JSON file 1", expiration=20)
    if ret != 1:
        raise Exception("set does not work")

    ret = r.set(resource_id + "_2", "JSON file 2", expiration=20)
    if ret != 1:
        raise Exception("set does not work")

    ret = r.set(resource_id + "_3", "JSON file 3", expiration=20)
    if ret != 1:
        raise Exception("set does not work")

    rlist = r.get_list(resource_id + "*")
    print("line 275", rlist)
    for resource in rlist:
        if "JSON file" not in resource:
            raise Exception("keys does not work")

    ret = r.set_termination(resource_id + "_1", expiration=5)
    if ret != 1:
        raise Exception("set_termination does not work")

    ret = r.set_termination(resource_id + "_2", expiration=5)
    if ret != 1:
        raise Exception("set_termination does not work")

    ret = r.set_termination(resource_id + "_3", expiration=5)
    if ret != 1:
        raise Exception("set_termination does not work")

    rlist = r.get_termination_list(resource_id + "*")
    print(rlist)
    if len(rlist) != 3:
        raise Exception("set_termination does not work")


if __name__ == "__main__":
    import os
    import signal
    import time

    pid = os.spawnl(
        os.P_NOWAIT, "/usr/bin/redis-server", "./redis.conf", "--port 7000"
    )

    time.sleep(1)

    try:
        r = RedisResourceInterface()
        r.connect(host="localhost", port=7000)
        test_resource_entries(r)
        r.disconnect()
    except Exception:
        raise
    finally:
        os.kill(pid, signal.SIGTERM)



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Raster map renderer

"""
from tempfile import NamedTemporaryFile
from actinia_core.processing.actinia_processing.ephemeral_processing import (
    EphemeralProcessing,
)
from actinia_core.processing.actinia_processing.ephemeral.base.renderer_base import (
    EphemeralRendererBase,
)

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class EphemeralRasterRenderer(EphemeralRendererBase):
    def __init__(self, *args):
        EphemeralRendererBase.__init__(self, *args)

    def _execute(self, skip_permission_check=True):
        """Render the raster image

        Workflow:

            1. The default region is set to the raster region
            2. User specific region settings are applied
            3. d.rast is invoked to create the PNG file

        """

        self._setup()

        raster_name = self.map_name
        options = self.rdc.user_data
        self.required_mapsets.append(self.mapset_name)

        with NamedTemporaryFile(suffix=".png") as file:
            result_file = file.name

        region_pc = self._setup_render_environment_and_region(
            options=options, result_file=result_file
        )

        pc = {}
        pc["1"] = {
            "module": "g.region",
            "inputs": {"raster": raster_name + "@" + self.mapset_name},
        }
        pc["2"] = region_pc
        pc["3"] = {
            "module": "d.rast",
            "inputs": {"map": raster_name + "@" + self.mapset_name},
            "flags": "n",
        }

        # Run the selected modules
        self.skip_region_check = True
        process_list = (
            self._create_temporary_grass_environment_and_process_list(
                process_chain=pc, skip_permission_check=True
            )
        )
        self._execute_process_list(process_list)

        self.module_results = result_file


class EphemeralRasterRGBRenderer(EphemeralRendererBase):
    def __init__(self, *args):
        EphemeralProcessing.__init__(self, *args)

    def _execute(self, skip_permission_check=True):
        """Render three raster layer as rgb image

        Workflow:

            1. The default region is set to the combined raster region
            2. User specific region settings are applied
            3. d.rgb is invoked to create the PNG file

        """

        self._setup()

        options = self.rdc.user_data
        self.required_mapsets.append(self.mapset_name)

        with NamedTemporaryFile(suffix=".png") as file:
            result_file = file.name

        region_pc = self._setup_render_environment_and_region(
            options=options, result_file=result_file
        )

        pc = {}
        pc["1"] = {
            "module": "g.region",
            "inputs": {
                "raster": "%s,%s,%s"
                % (options["red"], options["green"], options["blue"])
            },
        }
        pc["2"] = region_pc
        pc["3"] = {
            "module": "d.rgb",
            "inputs": {
                "red": options["red"],
                "green": options["green"],
                "blue": options["blue"],
            },
            "flags": "n",
        }

        self.request_data = pc

        # Run the selected modules
        self.skip_region_check = True
        process_list = (
            self._create_temporary_grass_environment_and_process_list(
                process_chain=pc, skip_permission_check=True
            )
        )
        self._execute_process_list(process_list)

        self.module_results = result_file


class EphemeralRasterShadeRenderer(EphemeralRendererBase):
    def __init__(self, *args):
        EphemeralRendererBase.__init__(self, *args)

    def _execute(self, skip_permission_check=True):
        """Render two raster layer as shaded image

        Workflow:

            1. The default region is set to the combined raster region
            2. User specific region settings are applied
            3. d.shade is invoked to create the PNG file

        """

        self._setup()

        options = self.rdc.user_data
        self.required_mapsets.append(self.mapset_name)

        with NamedTemporaryFile(suffix=".png") as file:
            result_file = file.name

        region_pc = self._setup_render_environment_and_region(
            options=options, result_file=result_file
        )

        pc = {}
        pc["1"] = {
            "module": "g.region",
            "inputs": {
                "raster": "%s,%s" % (options["shade"], options["color"])
            },
        }
        pc["2"] = region_pc
        pc["3"] = {
            "module": "d.shade",
            "inputs": {"shade": options["shade"], "color": options["color"]},
        }

        # Run the selected modules
        self.skip_region_check = True
        process_list = (
            self._create_temporary_grass_environment_and_process_list(
                process_chain=pc, skip_permission_check=True
            )
        )
        self._execute_process_list(process_list)

        self.module_results = result_file



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Download cache management

TODO: Tests required
"""

import os
from actinia_core.processing.actinia_processing.ephemeral.persistent_processing import (
    PersistentProcessing,
)
from actinia_core.core.common.process_object import Process
from actinia_core.core.common.exceptions import AsyncProcessError

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class DownloadCacheDelete(PersistentProcessing):
    """Delete the whole download cache directory"""

    def __init__(self, *args):
        PersistentProcessing.__init__(self, *args)
        self.user_download_cache_path = os.path.join(
            self.config.DOWNLOAD_CACHE, self.user_id
        )

    def _execute(self):
        self._setup()

        if os.path.exists(self.user_download_cache_path) and os.path.isdir(
            self.user_download_cache_path
        ):
            executable = "/bin/rm"
            args = ["-rf", self.user_download_cache_path]

            self._run_process(
                Process(
                    exec_type="exec",
                    executable=executable,
                    id="delete_download_cache_directory",
                    executable_params=args,
                )
            )

            os.mkdir(self.user_download_cache_path)
            self.finish_message = "Download cache successfully removed."
        else:
            raise AsyncProcessError(
                "Download cache directory <%s> does not exist."
                % self.user_download_cache_path
            )



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Tests: Download cache test case
"""
from flask.json import loads as json_load
import unittest
import os

try:
    from .test_resource_base import (
        ActiniaResourceTestCaseBase,
        global_config,
        URL_PREFIX,
    )
except ModuleNotFoundError:
    from test_resource_base import (
        ActiniaResourceTestCaseBase,
        global_config,
        URL_PREFIX,
    )

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Sören Gebbert"
__email__ = "soerengebbert@googlemail.com"


class DownloadCacheTestCase(ActiniaResourceTestCaseBase):
    def test_download_cache(self):
        # global_config.DOWNLOAD_CACHE = "/tmp/dcache_tmp"
        global_config.DOWNLOAD_CACHE_QUOTA = 1
        try:
            os.mkdir(global_config.DOWNLOAD_CACHE)
        except Exception:  # more precise exception gladly accepted
            pass

        try:
            os.mkdir(os.path.join(global_config.DOWNLOAD_CACHE, self.admin_id))
        except Exception:  # more precise exception gladly accepted
            pass

        rv = self.server.get(
            URL_PREFIX + "/download_cache", headers=self.admin_auth_header
        )
        print(rv.data.decode())
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        self.assertTrue("used" in json_load(rv.data)["process_results"])
        self.assertTrue("quota" in json_load(rv.data)["process_results"])
        self.assertTrue("free" in json_load(rv.data)["process_results"])
        self.assertTrue(
            "free_percent" in json_load(rv.data)["process_results"]
        )

        rv = self.server.delete(
            URL_PREFIX + "/download_cache", headers=self.admin_auth_header
        )
        print(rv.data.decode())
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        rv = self.server.get(
            URL_PREFIX + "/download_cache", headers=self.admin_auth_header
        )
        print(rv.data.decode())
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        self.assertTrue("used" in json_load(rv.data)["process_results"])
        self.assertTrue("quota" in json_load(rv.data)["process_results"])
        self.assertTrue("free" in json_load(rv.data)["process_results"])
        self.assertTrue(
            "free_percent" in json_load(rv.data)["process_results"]
        )

    # TODO: configure wrong download cache path differently (via config file)
    # see https://github.com/actinia-org/actinia-core/issues/422
    # def test_download_cache_error_1(self):
    #     if os.path.isdir("/tmp/dcache_tmp_nope") is True:
    #         os.rmdir("/tmp/dcache_tmp_nope")

    #     global_config.DOWNLOAD_CACHE = "/tmp/dcache_tmp_nope"
    #     global_config.DOWNLOAD_CACHE_QUOTA = 1

    #     rv = self.server.get(
    #         URL_PREFIX + "/download_cache", headers=self.admin_auth_header
    #     )
    #     print(rv.data.decode())
    #     self.assertEqual(
    #         rv.status_code,
    #         400,
    #         "HTML status code is wrong %i" % rv.status_code,
    #     )
    #     self.assertEqual(
    #         rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
    #     )

    # TODO: configure wrong download cache path differently (via config file)
    # see https://github.com/actinia-org/actinia-core/issues/422
    # def test_download_cache_error_2(self):
    #     if os.path.isdir("/tmp/dcache_tmp_nope") is True:
    #         os.rmdir("/tmp/dcache_tmp_nope")

    #     global_config.DOWNLOAD_CACHE = "/tmp/dcache_tmp_nope"
    #     global_config.DOWNLOAD_CACHE_QUOTA = 1

    #     rv = self.server.delete(
    #         URL_PREFIX + "/download_cache", headers=self.admin_auth_header
    #     )
    #     print(rv.data.decode())
    #     self.assertEqual(
    #         rv.status_code,
    #         400,
    #         "HTML status code is wrong %i" % rv.status_code,
    #     )
    #     self.assertEqual(
    #         rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
    #     )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""This module manages a raster layer resource as geotiff file that
will be created asynchronously.

The region settings of the mapset
or the raster layer region are used for export.
"""
from flask import jsonify, make_response
import pickle
from flask_restful_swagger_2 import swagger
from actinia_api.swagger2.actinia_core.apidocs import raster_export

from actinia_core.core.common.redis_interface import enqueue_job
from actinia_core.rest.base.endpoint_config import (
    check_endpoint,
    endpoint_decorator,
)
from actinia_core.rest.base.resource_base import ResourceBase
from actinia_core.processing.common.raster_export import start_job

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class AsyncEphemeralRasterLayerExporterResource(ResourceBase):
    """This class represents a raster layer resource as geotiff file that
    will be created asynchronously and streamed with the resource streamer.

    The region settings of the mapset is used for export.
    """

    def __init__(self):
        ResourceBase.__init__(self)

    @endpoint_decorator()
    @swagger.doc(check_endpoint("post", raster_export.post_doc))
    def post(self, location_name, mapset_name, raster_name):
        """Export an existing raster map layer as GeoTiff."""
        return self._execute(location_name, mapset_name, raster_name, False)

    def _execute(
        self, location_name, mapset_name, raster_name, use_raster_region
    ):
        rdc = self.preprocess(
            has_json=False,
            location_name=location_name,
            mapset_name=mapset_name,
            map_name=raster_name,
        )
        if rdc:
            rdc.set_user_data(use_raster_region)
            enqueue_job(self.job_timeout, start_job, rdc)

        html_code, response_model = pickle.loads(self.response_data)
        return make_response(jsonify(response_model), html_code)


class AsyncEphemeralRasterLayerRegionExporterResource(
    AsyncEphemeralRasterLayerExporterResource
):
    """This class represents a raster layer resource as geotiff file that
    will be created asynchronously.

    The region settings of the raster
    layer is used for export. A temporary mapset is used to modify
    the region settings.
    """

    def __init__(self):
        ResourceBase.__init__(self)

    @endpoint_decorator()
    @swagger.doc(check_endpoint("post", raster_export.region_post_doc))
    def post(self, location_name, mapset_name, raster_name):
        """Export an existing raster map layer as GeoTiff using the raster
        map specific region.
        """
        return self._execute(location_name, mapset_name, raster_name, True)



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Tests: Mapset test case
"""
from flask.json import loads as json_load
import unittest

try:
    from .test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX
except ModuleNotFoundError:
    from test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Sören Gebbert"
__email__ = "soerengebbert@googlemail.com"


class MapsetTestCase(ActiniaResourceTestCaseBase):
    def test_list_mapsets(self):
        rv = self.server.get(
            URL_PREFIX + "/locations/nc_spm_08/mapsets",
            headers=self.user_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        mapsets = json_load(rv.data)["process_results"]

        self.assertTrue("PERMANENT" in mapsets)
        self.assertTrue("user1" in mapsets)

    def test_mapsets_region_1(self):
        rv = self.server.get(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/PERMANENT/info",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        region_settings = json_load(rv.data)["process_results"]["region"]

        self.assertTrue("depths" in region_settings)
        self.assertTrue("ewres" in region_settings)
        self.assertTrue("cols" in region_settings)
        self.assertTrue("rows" in region_settings)

    def test_mapsets_region_2(self):
        rv = self.server.get(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/user1/info",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        region_settings = json_load(rv.data)["process_results"]["region"]

        self.assertTrue("depths" in region_settings)
        self.assertTrue("ewres" in region_settings)
        self.assertTrue("cols" in region_settings)
        self.assertTrue("rows" in region_settings)

    def test_mapset_creation_and_deletion(self):
        self.create_new_mapset("test_mapset")

        # Mapset already exists
        rv = self.server.post(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/test_mapset",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            400,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # Delete mapset
        rv = self.server.delete(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/test_mapset",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # Delete should fail, since mapset does not exists
        rv = self.server.delete(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/test_mapset",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            400,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

    def test_mapset_creation_and_deletion_unprivileged(self):
        # Create new mapsets as unprivileged user
        rv = self.server.post(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/test_mapset",
            headers=self.guest_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            401,
            "HTML status code is wrong %i" % rv.status_code,
        )

        # Delete mapset as unprivileged user
        rv = self.server.delete(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/test_mapset",
            headers=self.guest_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            401,
            "HTML status code is wrong %i" % rv.status_code,
        )

    def test_mapset_deletion_permanent_error(self):
        # Delete PERMANENT
        rv = self.server.delete(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/PERMANENT",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            400,
            "HTML status code is wrong %i" % rv.status_code,
        )

    def test_mapset_deletion_global_db_error(self):
        # Delete PERMANENT
        rv = self.server.delete(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/user1",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            400,
            "HTML status code is wrong %i" % rv.status_code,
        )

    def test_mapset_creation_and_locking(self):
        # Unlock mapset for deletion
        rv = self.server.delete(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/test_mapset_2/lock",
            headers=self.admin_auth_header,
        )
        print(rv.data)

        # Delete any existing mapsets
        rv = self.server.delete(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/test_mapset_2",
            headers=self.admin_auth_header,
        )
        print(rv.data)

        # Create new mapsets
        rv = self.server.post(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/test_mapset_2",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # Lock mapset
        rv = self.server.post(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/test_mapset_2/lock",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # get mapset lock(False)
        rv = self.server.get(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/test_mapset_2/lock",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        lock_status = json_load(rv.data)["process_results"]
        self.assertTrue(lock_status)

        # Unlock mapset
        rv = self.server.delete(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/test_mapset_2/lock",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # get mapset lock (False)
        rv = self.server.get(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/test_mapset_2/lock",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        lock_status = json_load(rv.data)["process_results"]
        self.assertFalse(lock_status)

        # Delete mapset
        rv = self.server.delete(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/test_mapset_2",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # get mapset lock (False)
        rv = self.server.get(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/test_mapset_2/lock",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            400,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # lock_status = json_load(rv.data)["process_results"]
        # self.assertFalse(lock_status)

        # Lock mapset
        # This locks the mapset even if it doesn't exist and even though an
        # error is logged. Skip until fixed. TODO reactivate
        # https://github.com/actinia-org/actinia-core/issues/487
        # rv = self.server.post(
        #     URL_PREFIX + "/locations/nc_spm_08/mapsets/test_mapset_2/lock",
        #     headers=self.admin_auth_header,
        # )
        # print(rv.data)
        # self.assertEqual(
        #     rv.status_code,
        #     400,
        #     "HTML status code is wrong %i" % rv.status_code,
        # )
        # self.assertEqual(
        #     rv.mimetype,
        #     "application/json",
        #     "Wrong mimetype %s" % rv.mimetype
        # )

        # Unlock mapset
        rv = self.server.delete(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/test_mapset_2/lock",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            400,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2023 mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
User management for no authentication
"""

# from passlib.apps import custom_app_context as pwd_context
# import jwt
# from datetime import datetime, timezone, timedelta
from actinia_core.core.common.config import global_config
from actinia_core.core.redis_user import redis_user_interface
from actinia_core.core.common.user import ActiniaUser


__author__ = "Anika Weinmann"
__copyright__ = "Copyright 2023, mundialis GmbH & Co. KG"
__maintainer__ = "mundialis GmbH & Co. KG"


class ActiniaUserNoAuth(ActiniaUser):
    """
    The Actinia Core user management class for no authentication
    """

    db = redis_user_interface

    def __init__(self):
        """Constructor
        Initialize and create a user object for no authentication.
        """
        self.user_id = global_config.DEFAULT_USER
        self.user_group = global_config.DEFAULT_USER_GROUP
        self.password_hash = None
        self.user_role = "superadmin"
        self.permissions = None
        self.cell_limit = global_config.MAX_CELL_LIMIT
        self.accessible_datasets = {
            "nc_spm_08": ["PERMANENT", "user1", "landsat"],
            "ECAD": ["PERMANENT"],
            "latlong_wgs84": ["PERMANENT"],
        }
        self.accessible_modules = global_config.MODULE_ALLOW_LIST
        self.process_num_limit = global_config.PROCESS_NUM_LIMIT
        self.process_time_limit = global_config.PROCESS_TIME_LIMT

    @staticmethod
    def create_user():
        """Create a new user object for no authentication and initialize it

        Returns:
            actinia_core_api.common.user_noauth.ActiniaUserNoAuth:
            A new user object in case of success, or None in case of failure
        """
        user = ActiniaUserNoAuth()
        user.hash_password("")

        if user.commit() is True:
            return user
        return None
